#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Person browser views.
"""
import urllib

from zope.authentication.interfaces import IAuthentication
from zope.interface import Interface
from zope.publisher.interfaces import NotFound
from zope.schema import Password
from zope.schema.interfaces import ValidationError
from zope.schema.interfaces import IIterableSource
from zope.schema.interfaces import ITitledTokenizedTerm
from zope.proxy import sameProxiedObjects
from zope.security import checkPermission
from zope.security.proxy import removeSecurityProxy
from zope.security.checker import canAccess
from zope.app.form.browser.source import SourceMultiCheckBoxWidget as SourceMultiCheckBoxWidget_
from zope.app.form.browser.interfaces import ITerms
from zope.app.dependable.interfaces import IDependable
from zope.publisher.browser import BrowserView
from zope.viewlet.interfaces import IViewletManager
from zope.component import getUtility
from zope.interface import implements
from zope.browserpage import ViewPageTemplateFile
from zope.viewlet.viewlet import ViewletBase
from zope.component import queryAdapter
from zope.component import adapts
from zope.catalog.interfaces import ICatalog
from zope.intid.interfaces import IIntIds
from zope.traversing.browser.absoluteurl import absoluteURL
from z3c.form import form, field, button, widget, term, validator
from z3c.form.browser.radio import RadioWidget
from zope.i18n import translate
from zope.i18n.interfaces.locales import ICollator

import schooltool.skin.flourish.page
import schooltool.skin.flourish.form
from schooltool.group.interfaces import IGroupContainer
from schooltool.common import SchoolToolMessage as _
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.catalog import buildQueryString
from schooltool.person.interfaces import IPasswordWriter
from schooltool.person.interfaces import IPerson, IPersonFactory
from schooltool.person.interfaces import IPersonPreferences
from schooltool.person.interfaces import IPersonContainer, IPersonContained
from schooltool import table
from schooltool.skin import flourish
from schooltool.securitypolicy.crowds import Crowd
from schooltool.securitypolicy.interfaces import ICrowd
from schooltool.table.column import IndexedLocaleAwareGetterColumn
from schooltool.table.table import url_cell_formatter


def SourceMultiCheckBoxWidget(field, request):
    source = field.value_type.source
    return SourceMultiCheckBoxWidget_(field, source, request)


class PersonPhotoView(BrowserView):
    """View that returns photo of a Person."""

    __used_for__ = IPerson

    def __call__(self):
        photo = self.context.photo
        if not photo:
            raise NotFound(self.context, u'photo', self.request)
        self.request.response.setHeader('Content-Type', "image/jpeg")
        return photo


class CalendarPeriodsWidgetTerms(term.BoolTerms):

    trueLabel = _('Show periods')
    falseLabel =  _('Show hours')


class CalendarPublicWidgetTerms(term.BoolTerms):

    trueLabel = _(u'the public')

    @property
    def falseLabel(self):
        person = self.context.__parent__
        return _('${person_full_name} and school administration',
                 mapping={'person_full_name': person.title})


class CalendarPeriodsRadioWidget(RadioWidget):

    @property
    def terms(self):
        return CalendarPeriodsWidgetTerms(self.context, self.request,
                                          self.form, self.field, self)


class CalendarPublicRadioWidget(RadioWidget):

    @property
    def terms(self):
        return CalendarPublicWidgetTerms(self.context, self.request,
                                         self.form, self.field, self)


def calendar_public_widget_label(adapter):
    person = adapter.context.__parent__
    return _("${person_full_name}'s calendar is visible to...",
             mapping={'person_full_name': person.title})


CalendarPublicWidgetLabel = widget.ComputedWidgetAttribute(
    calendar_public_widget_label,
    field=IPersonPreferences['cal_public'],
    widget=CalendarPublicRadioWidget)


def CalendarPeriodsWidgetFactory(field, request):
    return widget.FieldWidget(field, CalendarPeriodsRadioWidget(request))


def CalendarPublicWidgetFactory(field, request):
    return widget.FieldWidget(field, CalendarPublicRadioWidget(request))


class PersonPreferencesView(form.EditForm):
    """View used for editing person preferences."""

    fields = field.Fields(IPersonPreferences)
    fields['cal_periods'].widgetFactory = CalendarPeriodsWidgetFactory
    fields['cal_public'].widgetFactory = CalendarPublicWidgetFactory
    template = ViewPageTemplateFile('person_preferences.pt')

    @property
    def label(self):
        person = self.context.__parent__
        return _(u'Change preferences for ${person_full_name}',
                 mapping={'person_full_name': person.title})

    @button.buttonAndHandler(_("Apply"))
    def handle_edit_action(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        self.applyChanges(data)
        self.redirectToPerson()

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        self.redirectToPerson()

    def updateActions(self):
        super(PersonPreferencesView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def redirectToPerson(self):
        url = absoluteURL(self.context.__parent__, self.request)
        self.request.response.redirect(url)


class FlourishPasswordChangedView(flourish.form.DialogForm):

    dialog_submit_actions = ('ok',)
    label = None

    def initDialog(self):
        super(FlourishPasswordChangedView, self).initDialog()
        self.ajax_settings['dialog']['dialogClass'] = 'explicit-close-dialog'
        self.ajax_settings['dialog']['closeOnEscape'] = False

    @button.buttonAndHandler(_('ok-button', 'OK'), name='ok')
    def handle_submit_(self, action):
        app = ISchoolToolApplication(None)
        persons = app['persons']
        username = self.request.get('username', '')
        person = persons.get(username, None)
        if person is not None:
            url = absoluteURL(person, self.request)
        else:
            url = absoluteURL(app, self.request)
        self.request.response.redirect(url)
        self.ajax_settings['dialog'] = 'close'

    def updateActions(self):
        super(FlourishPasswordChangedView, self).updateActions()
        self.actions['ok'].addClass('button-ok')


class FlourishPersonPreferencesView(flourish.form.DialogForm,
                                    PersonPreferencesView):
    """View used for editing person preferences."""

    dialog_submit_actions = ('apply',)
    dialog_close_actions = ('cancel',)
    label = None

    @button.handler(PersonPreferencesView.buttons['apply'])
    def handleAdd(self, action):
        self.handleApply.func(self, action)
        # We never have errors, so just close the dialog.
        self.ajax_settings['dialog'] = 'close'
        # Also I assume the preferences don't change the parent
        # view content, so let's not reload it now.
        self.reload_parent = False

    def initDialog(self):
        super(FlourishPersonPreferencesView, self).initDialog()
        self.ajax_settings['dialog']['width'] = '306'
        self.ajax_settings['dialog']['dialogClass'] = 'narrow-dialog'


class FlourishPersonPreferencesLink(flourish.page.ModalFormLinkViewlet):

    @property
    def dialog_title(self):
        person = self.context
        title = _(u'Change preferences for ${person_full_name}',
                  mapping={'person_full_name': person.title})
        return translate(title, context=self.request)

    @property
    def enabled(self):
        prefs = IPersonPreferences(self.context)
        if not flourish.canEdit(prefs):
            return False
        return super(FlourishPersonPreferencesLink, self).enabled


class FlourishPersonDeleteView(flourish.form.DialogForm, form.EditForm):
    """View used for editing person preferences."""

    dialog_submit_actions = ('apply',)
    dialog_close_actions = ('cancel',)
    label = None

    @button.buttonAndHandler(_("Delete"), name='apply')
    def handleDelete(self, action):
        url = '%s/delete.html?delete.%s&CONFIRM' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.username)
        self.request.response.redirect(url)
        # We never have errors, so just close the dialog.
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        pass

    def updateActions(self):
        super(FlourishPersonDeleteView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class FlourishPersonDeleteLink(flourish.page.ModalFormLinkViewlet):

    @property
    def dialog_title(self):
        person = self.context
        title = _(u'Delete ${person_full_name}',
                  mapping={'person_full_name': person.title})
        return translate(title, context=self.request)

    def render(self, *args, **kw):
        dep = IDependable(removeSecurityProxy(self.context), None)
        if (dep is not None and dep.dependents()):
            return ''
        return flourish.page.ModalFormLinkViewlet.render(self, *args, **kw)


# Should this be moved to a interface.py file ?
class IGroupsSource(IIterableSource):
    pass


class GroupsSource(object):
    implements(IGroupsSource)

    def __iter__(self):
        return iter(ISchoolToolApplication(None)['groups'].values())

    def __len__(self):
        return len(ISchoolToolApplication(None)['groups'].values())


class GroupsTerm(object):
    implements(ITitledTokenizedTerm)

    def __init__(self, title, token, value):
        self.title = title
        self.token = token
        self.value = value


class GroupsTerms(object):
    implements(ITerms)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def getTerm(self, value):
        return GroupsTerm(value.title, value.__name__, value)

    def getValue(self, token):
        return ISchoolToolApplication(None)['groups'][token]


class IPasswordEditForm(Interface):
    """Schema for a person's edit form."""

    current = Password(
        title=_('Current password'),
        required=True)

    password = Password(
        title=_("New password"),
        required=True)

    verify_password = Password(
        title=_("Verify new password"),
        required=True)


class PersonPasswordEditView(form.Form):

    label = _("Edit password")
    template = ViewPageTemplateFile('password_form.pt')

    @property
    def fields(self):
        fields = field.Fields(IPasswordEditForm, ignoreContext=True)
        person = IPerson(self.request.principal)
        if person.username is not self.context.username:
            # Editing someone else's password
            fields = fields.omit('current')
        return fields

    @button.buttonAndHandler(_("Apply"))
    def handle_edit_action(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        writer = IPasswordWriter(self.context)
        writer.setPassword(data['password'])
        person = IPerson(self.request.principal)
        self.status = _('Password changed successfully')
        app = ISchoolToolApplication(None)
        url = '%s/password_changed.html?username=%s'
        url = url % (absoluteURL(app, self.request), self.context.username)
        self.dialog_show = True
        self.dialog_title = translate(self.status, context=self.request)
        self.dialog_url = url
        if person.username is self.context.username:
            auth = getUtility(IAuthentication)
            auth.setCredentials(self.request,
                                person.username,
                                data['password'])

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        # redirect to parent
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(PersonPasswordEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class FlourishPersonPasswordEditView(flourish.page.Page,
                                     PersonPasswordEditView):

    label = None
    legend = _('Change password')
    formErrorsMessage = _('Please correct the marked fields below.')

    def update(self):
        PersonPasswordEditView.update(self)


class WrongCurrentPassword(ValidationError):
    __doc__ = _('Wrong password supplied')


class PasswordsDontMatch(ValidationError):
    __doc__ = _('Supplied new passwords are not identical')


class CurrentPasswordValidator(validator.SimpleFieldValidator):

    def validate(self, value):
        super(CurrentPasswordValidator, self).validate(value)
        if value is not None and not self.context.checkPassword(value):
            raise WrongCurrentPassword(value)


validator.WidgetValidatorDiscriminators(CurrentPasswordValidator,
                                        field=IPasswordEditForm['current'])


class PasswordsMatchValidator(validator.SimpleFieldValidator):

    def validate(self, value):
        # XXX: hack to display the validation error next to the widget!
        super(PasswordsMatchValidator, self).validate(value)
        name = self.view.widgets['verify_password'].name
        verify = self.request.get(name)
        if value is not None and value not in (verify,):
            raise PasswordsDontMatch(value)


validator.WidgetValidatorDiscriminators(PasswordsMatchValidator,
                                        field=IPasswordEditForm['password'])


class IPersonInfoManager(IViewletManager):
    """Provides a viewlet hook for the information on a Person's page."""


class PasswordEditMenuItem(ViewletBase):
    """Viewlet that is visible if user can change the password of the context."""

    def render(self):
        if checkPermission('schooltool.edit', IPasswordWriter(self.context)):
            return super(PasswordEditMenuItem, self).render()


class FlourishPasswordLinkViewlet(flourish.page.LinkViewlet):

    def render(self):
        if checkPermission('schooltool.edit', IPasswordWriter(self.context)):
            return super(FlourishPasswordLinkViewlet, self).render()


class PersonFilterWidget(table.catalog.IndexedFilterWidget):
    """A filter widget for persons.

    Filters the list of available persons by title or groups they belong to.
    """

    template = ViewPageTemplateFile('person_filter.pt')

    def groupContainer(self):
        # XXX must know which group container to pick
        app = ISchoolToolApplication(None)
        return IGroupContainer(app, {})

    search_title_id = 'SEARCH_TITLE'
    search_group_id = 'SEARCH_GROUP'

    @property
    def parameters(self):
        return (self.search_title_id, self.search_group_id)

    def groups(self):
        groups = []
        container = self.groupContainer()
        collator = ICollator(self.request.locale)
        group_items = sorted(container.items(),
                             cmp=collator.cmp,
                             key=lambda (gid, g): g.title)
        for id, group in group_items:
            if len(group.members) > 0:
                groups.append({'id': id,
                               'title': "%s (%s)" % (group.title, len(group.members))})
        return groups

    @property
    def source(self):
        return self.context

    def filter(self, items):
        if 'CLEAR_SEARCH' in self.request:
            for parameter in self.parameters:
                self.request.form[parameter] = ''
            return items

        if self.search_group_id in self.request:
            group = self.groupContainer().get(self.request[self.search_group_id])
            if group:
                int_ids = getUtility(IIntIds)
                keys = set([int_ids.queryId(person)
                            for person in group.members])
                items = [item for item in items
                         if item['id'] in keys]

        if self.search_title_id in self.request:
            searchstr = self.request[self.search_title_id]
            query = buildQueryString(searchstr)
            if query:
                catalog = ICatalog(self.source)
                result = catalog['text'].apply(query)
                items = [item for item in items
                         if item['id'] in result]

        return items

    def active(self):
        for parameter in self.parameters:
            if parameter in self.request:
                return True
        return False

    def quote(self, param):
        return urllib.quote(unicode(param).encode('UTF-8'))

    def extra_url(self):
        url = ""
        for parameter in self.parameters:
            if parameter in self.request:
                url += '&%s=%s' % (parameter, self.quote(self.request.get(parameter)))
        return url


class PersonTableFilter(table.ajax.IndexedTableFilter,
                        PersonFilterWidget):

    template = ViewPageTemplateFile('f_person_table_filter.pt')
    title = _('First name, last name or username')

    @property
    def search_title_id(self):
        return self.manager.html_id+"-title"

    @property
    def search_group_id(self):
        return self.manager.html_id+"-group"

    @property
    def parameters(self):
        return (self.search_title_id, self.search_group_id)

    groupContainer = PersonFilterWidget.groupContainer
    groups = PersonFilterWidget.groups
    active = PersonFilterWidget.active
    extra_url = PersonFilterWidget.extra_url

    def filter(self, results):
        if self.ignoreRequest:
            return results
        return PersonFilterWidget.filter(self, results)


def person_table_columns():
    """Columns for the old person table.

    Functional tests expect the first column to be first name."""
    first_name = IndexedLocaleAwareGetterColumn(
        index='first_name',
        name='first_name',
        cell_formatter=url_cell_formatter,
        title=_(u'First Name'),
        getter=lambda i, f: i.first_name,
        subsort=True)
    last_name = IndexedLocaleAwareGetterColumn(
        index='last_name',
        name='last_name',
        cell_formatter=url_cell_formatter,
        title=_(u'Last Name'),
        getter=lambda i, f: i.last_name,
        subsort=True)
    return [first_name, last_name]


class PersonTable(table.ajax.IndexedTable):

    @property
    def source(self):
        return ISchoolToolApplication(None)['persons']

    def columns(self):
        return person_table_columns()

    def sortOn(self):
        return getUtility(IPersonFactory).sortOn()


class PersonTableFormatter(table.catalog.IndexedTableFormatter):
    """Person container specific table formatter."""

    def columns(self):
        return person_table_columns()

    def sortOn(self):
        return getUtility(IPersonFactory).sortOn()


class PersonView(BrowserView):
    """A Person info view."""

    __used_for__ = IPersonContained

    def __init__(self, context, request):
        BrowserView.__init__(self, context, request)


class HomeView(PersonView):
    """A Person's homepage."""

    @property
    def canModify(self):
        return canAccess(self.context.__parent__, '__delitem__')


class IPreferencesMenuViewlet(Interface):
    """Marker interface so we could use custom crowd for Preferences menu item"""


class PreferencesActionMenuViewlet(object):
    implements(IPreferencesMenuViewlet)


class PreferencesMenuViewletCrowd(Crowd):
    adapts(IPreferencesMenuViewlet)

    def contains(self, principal):
        """Returns true if you have the permission to see the calendar."""
        crowd = queryAdapter(IPersonPreferences(self.context.context),
                             ICrowd,
                             name="schooltool.view")
        return crowd.contains(principal)


class PersonAddPersonViewlet(object):

    @property
    def container(self):
        return IPersonContainer(self.context)

    @property
    def visible(self):
        if not checkPermission("schooltool.edit", self.container):
            return False
        authenticated = IPerson(self.request.principal, None)
        target = IPerson(self.context, None)
        if sameProxiedObjects(authenticated, target):
            return False
        return True


class FlourishPersonFilterWidget(PersonFilterWidget):

    template = ViewPageTemplateFile('f_person_filter.pt')

    def filter(self, items):
        if 'CLEAR_SEARCH' in self.request:
            for parameter in self.parameters:
                self.request.form[parameter] = ''
            return items

        if 'SEARCH_GROUP' in self.request:
            group = self.groupContainer().get(self.request['SEARCH_GROUP'])
            if group:
                int_ids = getUtility(IIntIds)
                keys = set([int_ids.queryId(person)
                            for person in group.members])
                items = [item for item in items
                         if item['id'] in keys]

        if 'SEARCH_TITLE' in self.request:
            search_title = self.request['SEARCH_TITLE']
            query = buildQueryString(search_title)
            if query:
                catalog = ICatalog(self.source)
                result = catalog['text'].apply(query)
                items = [item for item in items
                         if item['id'] in result]

        return items



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Main SchoolTool script.
"""
import os
import sys
import getopt
import locale
import gettext
import logging
import errno
import pkg_resources
from StringIO import StringIO
from collections import defaultdict

import ZConfig
import transaction
import zope.configuration.config
import zope.configuration.xmlconfig
from ZODB.ActivityMonitor import ActivityMonitor
from ZODB.interfaces import IDatabase
from zope.interface import directlyProvides, implements
from zope.component import provideUtility
from zope.component import provideAdapter, adapts
from zope.event import notify
from zope.publisher.interfaces.http import IHTTPRequest
from zope.i18n.interfaces import IUserPreferredLanguages
from zope.app.appsetup import DatabaseOpened
from zope.app.publication.zopepublication import ZopePublication
from zope.traversing.interfaces import IContainmentRoot
from zope.lifecycleevent import ObjectAddedEvent
from zope.app.dependable.interfaces import IDependable
from zope.component.hooks import getSite, setSite, setHooks
from zope.component import getUtility
from zope.component import getAdapters
from zope.interface import directlyProvidedBy
from zope.intid import IntIds
from zope.intid.interfaces import IIntIds
from zope.component.interfaces import ISite
from zope.site import LocalSiteManager

from schooltool.app.interfaces import ApplicationStartUpEvent
from schooltool.app.interfaces import ApplicationInitializationEvent
from schooltool.app.interfaces import IPluginInit, IPluginStartUp
from schooltool.app.interfaces import ICatalogStartUp
from schooltool.app.interfaces import ISchoolToolInitializationUtility
from schooltool.app.app import SchoolToolApplication
from schooltool.app.app import getApplicationPreferences
from schooltool.app import pdf
from schooltool.person.interfaces import IPersonFactory
from schooltool.app.interfaces import ICookieLanguageSelector
from schooltool.app.interfaces import CatalogSetUpEvent
from schooltool.app.interfaces import CatalogStartUpEvent
from schooltool.utility.utility import setUpUtilities
from schooltool.utility.utility import UtilitySpecification


MANAGER_USERNAME = 'manager'
MANAGER_PASSWORD = 'schooltool'

locale_charset = locale.getpreferredencoding()

localedir = os.path.join(os.path.dirname(__file__), '..', 'locales')
catalog = gettext.translation('schooltool', localedir, fallback=True)
_ = lambda us: catalog.ugettext(us).encode(locale_charset, 'replace')
_._domain = 'schooltool'


class Options(object):
    config = None
    pack = False
    restore_manager = False
    manager_password = MANAGER_PASSWORD
    config_file = ''


class CookieLanguageSelector(object):
    implements(ICookieLanguageSelector)

    def getLanguageList(self):
        return self.available_languages

    def getSelectedLanguage(self):
        return self.context.cookies.get("schooltool.lang", self.available_languages[0])

    def setSelectedLanguage(self, lang):
        self.context.response.setCookie("schooltool.lang", lang)


def setLanguage(lang):
    """Set the language for the user interface."""
    if lang == 'auto':
        return # language is negotiated at runtime through Accept-Language.

    if len(lang.split(",")) > 1:

        class CookiePreferredLanguage(CookieLanguageSelector):
            adapts(IHTTPRequest)
            implements(IUserPreferredLanguages)

            def __init__(self, context):
                self.context = context
                self.available_languages = [language.strip()
                                            for language in lang.split(",")]

            def getPreferredLanguages(self):
                return (self.getSelectedLanguage(),)

        provideAdapter(CookiePreferredLanguage, provides=IUserPreferredLanguages)
        return

    class SinglePreferredLanguage(object):

        adapts(IHTTPRequest)
        implements(IUserPreferredLanguages)

        def __init__(self, context):
            pass

        def getPreferredLanguages(self):
            return (lang, )

    # Replace the default adapter with one that always asks for the language
    # specified in the configuration file.
    provideAdapter(SinglePreferredLanguage)


class StreamWrapper(object):
    r"""Unicode-friendly wrapper for writable file-like objects.

    Here the terms 'encoding' and 'charset' are used interchangeably.

    The main use case for StreamWrapper is wrapping sys.stdout and sys.stderr
    so that you can forget worrying about charsets of your data.

        >>> import schooltool.app.main
        >>> old_locale_charset = schooltool.app.main.locale_charset
        >>> schooltool.app.main.locale_charset = 'UTF-8'

        >>> sw = StreamWrapper(StringIO())
        >>> print >> sw, u"Hello, world! \u00b7\u263b\u00b7"
        >>> sw.stm.getvalue()
        'Hello, world! \xc2\xb7\xe2\x98\xbb\xc2\xb7\n'

    By default printing Unicode strings to stdout/stderr will raise Unicode
    errors if the stream encoding does not include some characters you are
    printing.  StreamWrapper will replace unconvertable characters to question
    marks, therefore you should only use it for informative messages where such
    loss of information is acceptable.

        >>> schooltool.app.main.locale_charset = 'US-ASCII'
        >>> sw = StreamWrapper(StringIO())
        >>> print >> sw, u"Hello, world! \u00b7\u263b\u00b7"
        >>> sw.stm.getvalue()
        'Hello, world! ???\n'

    StreamWrapper converts all unicode strings that are written to it to the
    encoding defined in the wrapped stream's 'encoding' attribute, or, if that
    is None, to the locale encoding.  Typically the stream's encoding attribute
    is set when the stream is connected to a console device, and None when the
    stream is connected to a file.  On Unix systems the console encoding
    matches the locale charset, but on Win32 systems they differ.

        >>> s = StringIO()
        >>> s.encoding = 'ISO-8859-1'
        >>> sw = StreamWrapper(s)
        >>> print >> sw, u"Hello, world! \u00b7\u263b\u00b7"
        >>> sw.stm.getvalue()
        'Hello, world! \xb7?\xb7\n'

    You can print other kinds of objects:

        >>> sw = StreamWrapper(StringIO())
        >>> print >> sw, 1, 2,
        >>> print >> sw, 3
        >>> sw.stm.getvalue()
        '1 2 3\n'

    but not 8-bit strings:

        >>> sw = StreamWrapper(StringIO())
        >>> print >> sw, "\xff"
        Traceback (most recent call last):
          ...
        UnicodeDecodeError: 'ascii' codec can't decode byte 0xff in position 0: ordinal not in range(128)

    In addition to 'write', StreamWrapper provides 'flush' and 'writelines'

        >>> sw = StreamWrapper(StringIO())
        >>> sw.write('xyzzy\n')
        >>> sw.flush()
        >>> sw.writelines(['a', 'b', 'c', 'd'])
        >>> sw.stm.getvalue()
        'xyzzy\nabcd'

    Clean up:

        >>> schooltool.app.main.locale_charset = old_locale_charset

    """

    def __init__(self, stm):
        self.stm = stm
        self.encoding = getattr(stm, 'encoding', None)
        if self.encoding is None:
            self.encoding = locale_charset

    def write(self, obj):
        self.stm.write(obj.encode(self.encoding, 'replace'))

    def flush(self):
        self.stm.flush()

    def writelines(self, seq):
        for obj in seq:
            self.write(obj)

    def close(self):
        self.stm.close()


class UnicodeFileHandler(logging.StreamHandler):
    """A handler class which writes records to disk files.

    This class differs from logging.FileHandler in that it can handle Unicode
    strings with graceful degradation.
    """

    def __init__(self, filename):
        stm = StreamWrapper(open(filename, 'a'))
        logging.StreamHandler.__init__(self, stm)

    def close(self):
        self.stream.close()


def setUpLogger(name, filenames, format=None):
    """Set up a named logger.

    Sets up a named logger to log into filenames with the given format.
    Two filenames are special: 'STDOUT' means sys.stdout and 'STDERR'
    means sys.stderr.
    """
    formatter = logging.Formatter(format)
    logger = logging.getLogger(name)
    logger.propagate = False
    logger.setLevel(logging.INFO)
    for filename in filenames:
        if filename == 'STDOUT':
            handler = logging.StreamHandler(sys.stdout)
        elif filename == 'STDERR':
            handler = logging.StreamHandler(sys.stderr)
        else:
            handler = UnicodeFileHandler(filename)
        handler.setFormatter(formatter)
        logger.addHandler(handler)


class PluginDependency(object):

    def __init__(self, plugin, name,
                 other_plugin=None, other_name=None,
                 inverse=False):
        self.plugin, self.name = plugin, name
        self.other_plugin, self.other_name = other_plugin, other_name
        self.inverse = inverse

    def __str__(self):
        direction = '_before_'
        if self.inverse:
            direction = '_after_'
        if self.other_plugin:
            return ('%(plugin)s named "%(name)s" must be executed '
                    '%(before_or_after)s %(other)s named "%(other_name)s".' % {
                        'plugin': self.plugin,
                        'name': self.name,
                        'other': self.other_plugin,
                        'other_name': self.other_name,
                        'before_or_after': direction,
                        })
        return ('%(plugin)s named "%(name)s" must be executed '
                '%(before_or_after)s another.' % {
                    'plugin': self.plugin,
                    'name': self.name,
                    'before_or_after': direction,
                    })


class CyclicPluginActionOrderException(Exception):

    def __init__(self, dependencies):
        self.dependencies = dependencies

    def __repr__(self):
        return (_("Cannot resolve plugin action order:\n") +
                '\n'.join([str(dep) for dep in self.dependencies]))

    __str__ = __repr__


class PluginActionSorter(object):
    new, open, closed = 'new', 'open', 'closed'
    status = None
    result = None
    dependencies = None
    adapters = None

    def __init__(self, adapters):
        self.adapters = list(adapters)
        self.dependencies = defaultdict(lambda:list())
        cache = dict(self.adapters)
        for name, plugin in self.adapters:
            for after in list(plugin.after):
                self.dependencies[name].append(
                    PluginDependency(cache[after], after))
            for before in list(plugin.before):
                self.dependencies[before].append(
                    PluginDependency(plugin, name, inverse=True))

    def _visit(self, adapter):
        name, plugin = adapter
        if self.status[name] == self.new:
            self.status[name] = self.open
            for dependency in self.dependencies[name]:
                errors = self._visit((dependency.name, dependency.plugin))
                if errors is not None:
                    if not dependency.inverse:
                        return errors + [PluginDependency(
                            dependency.plugin, dependency.name,
                            plugin, name,
                            )]
                    else:
                        return errors + [PluginDependency(
                            plugin, name,
                            dependency.plugin, dependency.name,
                            inverse=True
                            )]
            self.result.append(plugin)
            self.status[name] = self.closed

        elif self.status[name] == self.closed:
            return None # already visited

        elif self.status[name] == self.open:
            return [] # a cyclic dependency

    def __call__(self):
        self.status = dict([(name, self.new)
                            for name, plugin in self.adapters])
        self.result = []
        for adapter in self.adapters:
            errors = self._visit(adapter)
            if errors is not None:
                raise CyclicPluginActionOrderException(errors)
        return self.result


def executePluginActions(actions):
    for action in actions:
        try:
            action()
        except Exception, exception:
            print >> sys.stderr, "Failed to execute %s: %s" % (action, exception)


def startSchoolToolCatalogs(event):
    adapters = getAdapters((event.object, ), ICatalogStartUp)
    sorter = PluginActionSorter(adapters)
    executePluginActions(sorter())


def initializeSchoolToolPlugins(event):
    adapters = getAdapters((event.object, ), IPluginInit)
    sorter = PluginActionSorter(adapters)
    executePluginActions(sorter())


def startSchoolToolPlugins(event):
    adapters = getAdapters((event.object, ), IPluginStartUp)
    sorter = PluginActionSorter(adapters)
    executePluginActions(sorter())


def get_schooltool_plugin_configurations():
    """Returns a list of tuples (configuration, handler).

    Configuration is a string containing XML that will be included in
    the config-schema.

    Handler is a function that performs the actual work based on the
    configuration.
    """

    plugin_configurations = list(pkg_resources.iter_entry_points(
            'schooltool.plugin_configuration'))
    return [entry.load().get_configuration()
            for entry in plugin_configurations]


plugin_configurations = get_schooltool_plugin_configurations()


class SchoolToolMachinery(object):

    ZCONFIG_SCHEMA = os.path.join(os.path.dirname(__file__),
                                  'config-schema.xml')

    Options = Options

    def readConfig(self, filename):
        # Read configuration file
        schema_string = open(self.ZCONFIG_SCHEMA).read()
        plugins = [configuration
                   for (configuration, handler) in plugin_configurations]
        schema_string = schema_string % {'plugins': "\n".join(plugins)}
        schema_file = StringIO(schema_string)

        schema = ZConfig.loadSchemaFile(schema_file, self.ZCONFIG_SCHEMA)

        config, handler = ZConfig.loadConfig(schema, filename)
        return config, handler

    def configureComponents(self, options, site_zcml=None):
        """Configure Zope 3 components."""
        # Hook up custom component architecture calls
        setHooks()
        context = zope.configuration.config.ConfigurationMachine()

        for config, handler in plugin_configurations:
            if handler is not None:
                handler(options, context)

        if options.config.devmode:
            context.provideFeature('devmode')

        zope.configuration.xmlconfig.registerCommonDirectives(context)

        if site_zcml is None:
            site_zcml = options.config.site_definition
        context = zope.configuration.xmlconfig.file(
            site_zcml, context=context)

        # Store the configuration context
        from zope.app.appsetup import appsetup
        appsetup.__dict__['__config_context'] = context

    def configureReportlab(self, fontdirs):
        """Configure reportlab given a path to TrueType fonts.

        Disables PDF support in SchoolTool if fontdir is empty.
        Outputs a warning to stderr in case of errors.
        """
        if not fontdirs:
            return

        existing_directories = []
        for fontdir in fontdirs.split(':'):
            if os.path.isdir(fontdir):
                existing_directories.append(fontdir)

        if not existing_directories:
            print >> sys.stderr, (_("Warning: font directories '%s' do"
                                    " not exist.\nPDF support disabled.")
                                  % fontdirs)
            return

        for font_file in pdf.font_map.values():
            font_exists = False
            for fontdir in existing_directories:
                font_path = os.path.join(fontdir, font_file)
                if os.path.exists(font_path):
                    font_exists = True
            if not font_exists:
                print >> sys.stderr, _("Warning: font '%s' does not exist"
                                       " in the font directories '%s'.\n"
                                       "PDF support disabled.") % (font_file,
                                                                   fontdirs)
                return

        pdf.setUpFonts(existing_directories)

    def configure(self, options):
        self.options = options
        self.configureComponents(
            options,
            site_zcml=options.config.site_definition)
        setLanguage(options.config.lang)
        self.configureReportlab(options.config.reportlab_fontdir)


class SchoolToolServer(SchoolToolMachinery):

    system_name = "SchoolTool"

    def load_options(self, argv):
        """Parse the command line and read the configuration file."""
        options = self.Options()

        # Parse command line
        progname = os.path.basename(argv[0])
        try:
            opts, args = getopt.gnu_getopt(argv[1:], 'c:hdr:',
                                           ['config=', 'pack',
                                            'help',
                                            'restore-manager=',
                                            'manage'])
        except getopt.error, e:
            print >> sys.stderr, "%s: %s" % (progname, e)
            print >> sys.stderr, _("Run %s -h for help.") % progname
            sys.exit(1)
        for k, v in opts:
            if k in ('-h', '--help'):
                print (_("\n"
                        "Usage: %s [options]\n"
                        "Options:\n"
                        "  -c, --config xxx       use this configuration file\n"
                        "  -h, --help             show this help message\n"
                        "  --pack                 pack the database\n"
                        "  -r, --restore-manager password\n"
                        "                         restore the manager user with the provided password\n"
                        "                         (read password from the standart input if 'password'\n"
                        "                         is '-')\n"
                        ) % progname).strip()
                sys.exit(0)
            if k in ('-c', '--config'):
                options.config_file = v
            if k in ('-p', '--pack'):
                options.pack = True
            if k in ('-r', '--restore-manager'):
                options.restore_manager = True
                if v != '-':
                    options.manager_password = v
                else:
                    from getpass import getpass
                    password = getpass(_("Manager password: "))
                    options.manager_password = password

        if not options.config_file:
            print >> sys.stderr, _("No configuration file given")
            sys.exit(1)

        print _("Reading configuration from %s") % options.config_file
        try:
            options.config, handler = self.readConfig(options.config_file)
        except ZConfig.ConfigurationError, e:
            print >> sys.stderr, "%s: %s" % (progname, e)
            sys.exit(1)
        if options.config.database.config.storage is None:
            print >> sys.stderr, "%s: %s" % (progname, _(
                "No storage defined in the configuration file."))

            sys.exit(1)

        return options

    def bootstrapSchoolTool(self, db):
        """Bootstrap SchoolTool database."""
        connection = db.open()
        root = connection.root()
        app_obj = root.get(ZopePublication.root_name)
        if app_obj is None:
            app = SchoolToolApplication()

            # Run school specific initialization code
            initializationUtility = getUtility(
                ISchoolToolInitializationUtility)
            initializationUtility.initializeApplication(app)

            directlyProvides(app, directlyProvidedBy(app) + IContainmentRoot)
            root[ZopePublication.root_name] = app
            # savepoint to make sure that the app object has
            # a _p_jar. This is needed to make things like
            # KeyReference work, which in turn is needed to
            # make the catalog work. We make this savepoint
            # optimistic because it will then work with any registered
            # data managers that do not support this feature.
            transaction.savepoint(optimistic=True)

            # set up the site so that local utility setups and catalog
            # indexing would work properly
            if not ISite.providedBy(app):
                app.setSiteManager(LocalSiteManager(app))

            setSite(app)

            # We must set up the int ids utility before setting up any
            # of the plugin specific catalogs, as catalogs rely on
            # IntIds being present
            setUpUtilities(app, [UtilitySpecification(IntIds, IIntIds)])

            # tell plugins to initialize their catalogs, must be done
            # before initializing plugins themselves or else all the
            # initial groups, persons, resources will not get indexed
            notify(CatalogSetUpEvent(app))
            notify(CatalogStartUpEvent(app))

            # initialize plugins themselves
            notify(ApplicationInitializationEvent(app))

            notify(ObjectAddedEvent(app))

            # unset the site so we would not confuse other
            # bootstraping code
            setSite(None)

            self.restoreManagerUser(app, MANAGER_PASSWORD)
            self.initializePreferences(app)
        transaction.commit()
        connection.close()

    def initializePreferences(self, app):
        prefs = getApplicationPreferences(app)
        prefs.initializeTimezone()

    def startApplication(self, db):
        last_site = getSite()
        connection = db.open()
        root = connection.root()
        app = root[ZopePublication.root_name]
        setSite(app)
        notify(CatalogStartUpEvent(app))
        notify(ApplicationStartUpEvent(app))
        setSite(last_site)
        transaction.commit()
        connection.close()

    def restoreManagerUser(self, app, password):
        """Ensure there is a manager user

        Create a user if needed, set password to default, grant
        manager permissions
        """
        # set the site so that catalog utilities and person factory
        # utilities and subscribers were available
        setSite(app)
        if MANAGER_USERNAME not in app['persons']:
            factory = getUtility(IPersonFactory)
            manager = factory.createManagerUser(MANAGER_USERNAME)
            app['persons'][MANAGER_USERNAME] = manager
            IDependable(manager).addDependent('')
        manager = app['persons'][MANAGER_USERNAME]
        manager.setPassword(password)
        app['persons'].super_user = manager
        setSite(None)

    def setup(self, options):
        """Configure SchoolTool."""
        setUpLogger(None, options.config.error_log_file,
                    "%(asctime)s %(message)s")
        setUpLogger('accesslog', options.config.web_access_log_file)

        # Shut up ZODB lock_file, because it logs tracebacks when unable
        # to lock the database file, and we don't want that.
        logging.getLogger('ZODB.lock_file').disabled = True

        self.configure(options)

        # Connect to the database
        db = self.openDB(options)

        return db

    def openDB(self, options):
        # Open the database
        db_configuration = options.config.database
        try:
            db = db_configuration.open()
            if options.pack:
                db.pack()
        except IOError, e:
            print >> sys.stderr, _("Could not initialize the database:\n%s") % e
            if e.errno == errno.EAGAIN: # Resource temporarily unavailable
                print >> sys.stderr, _("\nPerhaps another %s instance"
                                       " is using it?") % self.system_name
            sys.exit(1)

        self.bootstrapSchoolTool(db)
        notify(DatabaseOpened(db))

        if options.restore_manager:
            connection = db.open()
            root = connection.root()
            app = root[ZopePublication.root_name]
            self.restoreManagerUser(app, options.manager_password)
            transaction.commit()
            connection.close()

        # TODO: don't start application when running from command line
        #if options.pack or options.restore_manager:
        #    return db

        self.startApplication(db)

        provideUtility(db, IDatabase)
        db.setActivityMonitor(ActivityMonitor())

        return db

    def main(self, argv=sys.argv):
        options = self.load_options(argv)
        db = self.setup(options)


def main():
    SchoolToolServer().main()

if __name__ == '__main__':
    main()



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for SchoolTool PDF support.
"""

import os
import sys
import unittest
import doctest

from schooltool.app.pdf import setUpFonts


def doctest_setUpFonts():
    r"""TrueType font setup tests.

        >>> from schooltool.app import pdf

    The actual setup has been done at import-time by the test_suite function.
    We only test the results here.

    Let's check that the TrueType fonts have been configured:

        >>> from reportlab.pdfbase import pdfmetrics

        >>> pdfmetrics.getFont('Times_New_Roman')
        <reportlab.pdfbase.ttfonts.TTFont instance at ...>
        >>> pdfmetrics.getFont('Times_New_Roman_Bold')
        <reportlab.pdfbase.ttfonts.TTFont instance at ...>
        >>> pdfmetrics.getFont('Times_New_Roman_Italic')
        <reportlab.pdfbase.ttfonts.TTFont instance at ...>
        >>> pdfmetrics.getFont('Times_New_Roman_Bold_Italic')
        <reportlab.pdfbase.ttfonts.TTFont instance at ...>

        >>> pdfmetrics.getFont('Arial_Normal')
        <reportlab.pdfbase.ttfonts.TTFont instance at ...>
        >>> pdfmetrics.getFont('Arial_Bold')
        <reportlab.pdfbase.ttfonts.TTFont instance at ...>
        >>> pdfmetrics.getFont('Arial_Italic')
        <reportlab.pdfbase.ttfonts.TTFont instance at ...>
        >>> pdfmetrics.getFont('Arial_Bold_Italic')
        <reportlab.pdfbase.ttfonts.TTFont instance at ...>

    For our Serif font (normal paragraphs), the bold/italic mappings
    are registered:

        >>> from reportlab.lib.fonts import tt2ps, ps2tt

        >>> tt2ps('Times_New_Roman', 0, 0)
        'Times_New_Roman'
        >>> tt2ps('Times_New_Roman', 1, 0)
        'Times_New_Roman_Bold'
        >>> tt2ps('Times_New_Roman', 0, 1)
        'Times_New_Roman_Italic'
        >>> tt2ps('Times_New_Roman', 1, 1)
        'Times_New_Roman_Bold_Italic'

        >>> ps2tt('Times_New_Roman')
        ('times_new_roman', 0, 0)
        >>> ps2tt('Times_New_Roman_Bold')
        ('times_new_roman', 1, 0)
        >>> ps2tt('Times_New_Roman_Italic')
        ('times_new_roman', 0, 1)
        >>> ps2tt('Times_New_Roman_Bold_Italic')
        ('times_new_roman', 1, 1)

        >>> tt2ps('Arial_Normal', 0, 0)
        'Arial_Normal'
        >>> tt2ps('Arial_Normal', 1, 0)
        'Arial_Bold'
        >>> tt2ps('Arial_Normal', 0, 1)
        'Arial_Italic'
        >>> tt2ps('Arial_Normal', 1, 1)
        'Arial_Bold_Italic'

        >>> ps2tt('Arial_Normal')
        ('arial_normal', 0, 0)
        >>> ps2tt('Arial_Bold')
        ('arial_normal', 1, 0)
        >>> ps2tt('Arial_Italic')
        ('arial_normal', 0, 1)
        >>> ps2tt('Arial_Bold_Italic')
        ('arial_normal', 1, 1)

# XXX This part fails on Windows, I don't yet know why.
#    If the fonts can not be found, setUpFonts() will
#    raise an exception:
#
#        >>> import reportlab.rl_config
#        >>> real_path = reportlab.rl_config.TTFSearchPath[-1]
#        >>> del reportlab.rl_config.TTFSearchPath[-1]
#
#        >>> pdf.setUpFonts('/definitely/nonexistent')
#        Traceback (most recent call last):
#          ...
#        TTFError: Can't open file "....ttf"
#
#    Clean up:
#
#        >>> reportlab.rl_config.TTFSearchPath.append(real_path)

    """


def tryToSetUpReportLab():
    """Try to set up reportlab.

    Returns True without doing anything if pdf.isEnabled() is True.

    Tries to guess the location of fonts.  Returns True on success,
    False if reportlab is not available or fonts could not be found.

    If something breaks during setUpFonts, the exception
    will be propagated up.
    """
    try:
        import reportlab
    except ImportError:
        return False # We don't have reportlab, so we can't get anywhere.

    from schooltool.app import pdf
    if pdf.isEnabled():
        return True # Assume that reportlab has been configured already.

    # Heuristic to try and find the TrueType fonts.
    font_dirs = ['/usr/share/fonts/truetype/ttf-liberation', # Debian
                 '/usr/share/fonts/truetype/liberation', # Fedora
                 '/usr/share/fonts/truetype/ubuntu-font-family', # Ubuntu font family
                 r'C:\WINDOWS\Fonts']
    try:
        setUpFonts(font_dirs)
        return True
    except reportlab.pdfbase.ttfonts.TTFError:
        return False


def test_suite():
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF
                   | doctest.REPORT_ONLY_FIRST_FAILURE
                   | doctest.NORMALIZE_WHITESPACE)
    suite = unittest.TestSuite()

    success = tryToSetUpReportLab()
    if success:
        suite.addTest(doctest.DocTestSuite(optionflags=optionflags))
    else:
        print >> sys.stderr, ("reportlab or TrueType fonts not found;"
                              " PDF generator tests skipped")

    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.timetable.browser.calendar
"""
import doctest
import pytz
import unittest
from pprint import pprint
import  datetime

from schooltool.common import parse_time_range, format_time_range
from schooltool.timetable.browser.calendar import ScheduleDailyCalendarRowsView
from schooltool.timetable.schedule import iterMeetingsInTimezone


class MeetingStub(object):

    def __init__(self,
                 time="00:00-00:15",
                 year=2011, month=10, day=30,
                 timezone="UTC"):
        timezone = pytz.timezone(timezone)
        timestart, duration = parse_time_range(time)
        self.dtstart = timezone.localize(datetime.datetime(
            year, month, day,
            timestart.hour, timestart.minute))
        self.duration = duration


class ScheduleStub(object):
    timezone = "UTC"

    def __init__(self, meetings=None):
        if meetings:
            self.timezone = meetings[0].dtstart.tzinfo.zone
        self.meetings = meetings and list(meetings) or []

    def iterMeetings(self, start_date, until_date=None):
        if until_date is None:
            until_date = start_date
        for m in self.meetings:
            d = m.dtstart.date()
            if d >= start_date and d <= until_date:
                yield m


def test_ScheduleDailyCalendarRowsView_calendarRows():
    """Tests for ImmutableScheduleCalendar_calendarRows.

        >>> view = ScheduleDailyCalendarRowsView('context', 'request')

        >>> view._schedule = ScheduleStub([
        ...    MeetingStub("10:00-10:15"), MeetingStub("10:20-10:45"),
        ...    MeetingStub("10:55-12:30"), MeetingStub("14:00-14:55"),
        ...    ])

        >>> def getDefaultMeetings(d, tz):
        ...     return list(iterMeetingsInTimezone(view._schedule, tz, d))

        >>> view.getDefaultMeetings = getDefaultMeetings
        >>> view._timezone = 'UTC'
        >>> view.getPersonTimezone = lambda *a: pytz.timezone(view._timezone)
        >>> view.meetingTitle = lambda *a: 'meeting'
        >>> view.rowTitle = lambda *a: 'row'

        >>> def printRows(rows):
        ...     for r in rows:
        ...         dts = r[1].astimezone(pytz.timezone(view._timezone))
        ...         print '%-7s %s' % (r[0], format_time_range(dts.time(), r[2]))

    calendarRows generates rows for every hour in given range + 1 hour,
    inserting scheduled meetings where appropriate.

        >>> printRows(view.calendarRows(datetime.date(2011, 10, 30),
        ...                             8, 17, 'dummy'))
        row     08:00-09:00
        row     09:00-10:00
        meeting 10:00-10:15
        row     10:15-10:20
        meeting 10:20-10:45
        row     10:45-10:55
        meeting 10:55-12:30
        row     12:30-13:00
        row     13:00-14:00
        meeting 14:00-14:55
        row     14:55-16:00
        row     16:00-17:00
        row     17:00-18:00

    It works properly when displayed in timezones other than source meetings.

        >>> view._timezone = 'Europe/Vilnius'

        >>> printRows(view.calendarRows(datetime.date(2011, 10, 30),
        ...                             11, 17, 'dummy'))
        row     11:00-12:00
        meeting 12:00-12:15
        row     12:15-12:20
        meeting 12:20-12:45
        row     12:45-12:55
        meeting 12:55-14:30
        row     14:30-15:00
        row     15:00-16:00
        meeting 16:00-16:55
        row     16:55-18:00

    If we ask to display a time range, but there are earlier or later meetings,
    date range is streched to accomodate them.

        >>> printRows(view.calendarRows(datetime.date(2011, 10, 30),
        ...                             13, 14, 'dummy'))
        meeting 12:00-12:15
        row     12:15-12:20
        meeting 12:20-12:45
        row     12:45-12:55
        meeting 12:55-14:30
        row     14:30-15:00
        row     15:00-16:00
        meeting 16:00-16:55

    If there are meetings during DST change, this will be reflected in
    calendar rows.  When clocks are adjusted backwards, the doubled hour
    is properly displayed as separate entries.

        >>> view._schedule = ScheduleStub([
        ...    MeetingStub("23:00-23:15", day=29),
        ...    MeetingStub("00:00-00:15"), MeetingStub("01:00-01:15"),
        ...    ])

        >>> printRows(view.calendarRows(datetime.date(2011, 10, 30),
        ...                             1, 4, 'dummy'))
        row     01:00-02:00
        meeting 02:00-02:15
        row     02:15-03:00
        meeting 03:00-03:15
        row     03:15-04:00
        meeting 03:00-03:15
        row     03:15-04:00
        row     04:00-05:00

    Let's look at DST shift in a day without meetings:

        >>> view._schedule = ScheduleStub([])

        >>> printRows(view.calendarRows(datetime.date(2011, 10, 30),
        ...                             2, 4, 'dummy'))
        row     02:00-03:00
        row     03:00-04:00
        row     03:00-04:00
        row     04:00-05:00

        >>> printRows(view.calendarRows(datetime.date(2011, 10, 30),
        ...                             3, 5, 'dummy'))
        row     03:00-04:00
        row     04:00-05:00
        row     05:00-06:00

    Rows are limited to a given date in viewed timezone.

        >>> view._timezone = 'UTC'
        >>> view._schedule = ScheduleStub([
        ...    MeetingStub("23:00-00:00", day=29),
        ...    MeetingStub("23:00-23:30", day=30),
        ...    MeetingStub("23:30-00:00", day=30),
        ...    MeetingStub("00:00-01:00", day=31),
        ...    ])

        >>> printRows(view.calendarRows(datetime.date(2011, 10, 30),
        ...                             22, 22, 'dummy'))
        row     22:00-23:00
        meeting 23:00-23:30
        meeting 23:30-00:00

    Views may pass 24 as the end hour.

        >>> printRows(view.calendarRows(datetime.date(2011, 10, 30),
        ...                             22, 24, 'dummy'))
        row     22:00-23:00
        meeting 23:00-23:30
        meeting 23:30-00:00

    """


def setUp(test=None):
    pass


def tearDown(test=None):
    pass


def test_suite():
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF |
                   doctest.NORMALIZE_WHITESPACE)
    return unittest.TestSuite([
            doctest.DocTestSuite(optionflags=optionflags,
                                 setUp=setUp, tearDown=tearDown),
           ])



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool skin.
"""

from zope.component import getAdapter, subscribers
from zope.interface import Interface, implements
from zope.schema import Object
from zope.i18n.interfaces import IUserPreferredLanguages
from zope.publisher.interfaces.browser import IDefaultBrowserLayer
from zope.viewlet.interfaces import IViewletManager
from zope.viewlet.manager import ViewletManagerBase
from zope.viewlet.metadirectives import IViewletDirective
from zope.configuration.fields import MessageID
from zope.publisher.browser import BrowserView
from zope.browsermenu.menu import getMenu
from zope.traversing.browser.absoluteurl import absoluteURL

from z3c.form.interfaces import IFormLayer
from z3c.formui.interfaces import ICSS as IZ3cFormCSS

from schooltool.app.interfaces import ICookieLanguageSelector
from schooltool.securitypolicy.crowds import Crowd
from schooltool.securitypolicy.interfaces import ICrowd
from schooltool.app.interfaces import ISchoolToolApplication


class IJavaScriptManager(IViewletManager):
    """Provides a viewlet hook for the javascript link entries."""


class ICSSManager(IZ3cFormCSS):
    """Provides a viewlet hook for the CSS link entries."""


class IHeaderManager(IViewletManager):
    """Provides a viewlet hook for the header of a page."""


class INavigationManager(IViewletManager):
    """Provides a viewlet hook for the navigation section of a page."""


class IActionMenuManager(IViewletManager):
    """Provides a viewlet hook for the action menu."""

    def title():
        """Title that will be displayed before the action menu or submenu."""

    def subItems():
        """Sub items that will have their own menus in the main action menu.

        Returns a list of ISubMenuItem objects.
        """


class IMenuBarMenuManager(IViewletManager):
    """Provides a viewlet hook for the entire menu bar."""


class IActionMenuContext(Interface):
    """Schema for attributes required by ActionMenuViewletManager."""

    target = Object(Interface)


class OrderedViewletManager(ViewletManagerBase):
    """Viewlet manager that orders viewlets by their 'order' attribute.

    The order attribute can be a string, but it will be sorted numerically
    (i.e. '1' before '5' before '20').  The attribute is optional; viewlets
    without an ``order`` attribute will be sorted alphabetically by their
    ``title`` attribute, and placed below all the ordered viewlets.
    """

    def sort(self, viewlets):
        """Sort the viewlets.

        ``viewlets`` is a list of tuples of the form (name, viewlet).
        """

        def key_func((name, viewlet)):
            if hasattr(viewlet, 'order'):
                return (0, int(viewlet.order))
            elif hasattr(viewlet, 'title'):
                return (1, viewlet.title)
            else:
                raise AttributeError('%r viewlet has neither order nor title'
                                     % name)

        return sorted(viewlets, key=key_func)

    def filter(self, viewlets):
        viewlets = ViewletManagerBase.filter(self, viewlets)
        return [(name, viewlet) for (name, viewlet) in viewlets
                if not IDisableViewlet.providedBy(viewlet)]

    def render(self):
        """See zope.contentprovider.interfaces.IContentProvider"""
        # Now render the view
        viewlets_with_content = []
        for viewlet in self.viewlets:
            content = viewlet.render()
            if content and content.strip():
                viewlets_with_content.append(viewlet)

        # XXX we render the content twice, hope it's not affecting
        # anything
        if self.template:
            return self.template(viewlets=viewlets_with_content)
        else:
            return u'\n'.join([viewlet.render()
                               for viewlet in viewlets_with_content])


class NavigationViewlet(object):
    """A navigation viewlet base class."""

    def actualContext(self):
        return ISchoolToolApplication(None)

    def appURL(self):
        return absoluteURL(ISchoolToolApplication(None), self.request)


class TopLevelContainerNavigationViewlet(NavigationViewlet):
    """A base class of navigation viewlet for top level containers."""

    def actualContext(self):
        """Actual context is the container this viewlet links to."""
        return ISchoolToolApplication(None)[self.link]


class ISubMenuItem(Interface):
    """Marker interface for objects that will be displayed as items in submenus.
    """


class MenuBarMenuViewletManager(OrderedViewletManager):
    """Viewlet manager for displaying the various menus at the top of a page."""

    implements(IMenuBarMenuManager)


class ActionMenuViewletManager(OrderedViewletManager):
    """Viewlet manager for displaying the action menu."""

    implements(IActionMenuContext, IActionMenuManager)

    def title(self):
        return ''

    def getSubItems(self, context):
        """Collect all items that should be displayed in the submenu of context.

        Submenu items are registered as subscribers with interface
        ISubMenuItem.
        """
        return list(subscribers((context, ), ISubMenuItem))

    def update(self):
        # set the right context for menu items that will be set up by
        # OrderedViewletManager.update
        self.context = self.target

        # We could just check for ISubMenuItem on self.context, though
        # that would be less reliable.
        sub_items = self.getSubItems(self.context.__parent__)
        in_submenu_of_parent = self.context in sub_items
        displayed_in_submenu = IActionMenuManager.providedBy(self.__parent__)
        if (in_submenu_of_parent and not displayed_in_submenu):
            # display menu as if looking at the parent object
            self.context = self.context.__parent__
        self.orderedViewletManagerUpdate()

    def orderedViewletManagerUpdate(self):
        OrderedViewletManager.update(self)

    def subItems(self):
        """See zope.contentprovider.interfaces.IContentProvider"""
        return self.getSubItems(self.context)


class ActionMenuViewlet(object):

    def getMenu(self):
        return getMenu(self.menu, self.context, self.request)


class NavigationViewletViewCrowd(Crowd):
    """A crowd for navigation viewlets.

    Checks permissions on the actual context rather than the parent. Parents
    of viewlets are pretty useless in this case as NavigationViewlets
    are used in a global navigation menu.
    """

    def contains(self, principal):
        context = self.context.actualContext()
        crowd = getAdapter(context, ICrowd, name='schooltool.view')
        return crowd.contains(principal)


class NavigationViewletEditCrowd(Crowd):
    """A crowd for navigation viewlets."""

    def contains(self, principal):
        context = self.context.actualContext()
        crowd = getAdapter(context, ICrowd, name='schooltool.edit')
        return crowd.contains(principal)


class ISchoolToolLayer(IDefaultBrowserLayer, IFormLayer):
    """SchoolTool layer."""


class ISchoolToolSkin(ISchoolToolLayer):
    """The same SchoolTool layer. BBB naming."""


class LanguageSelectorViewlet(BrowserView):

    def selected_lang(self):
        upl = IUserPreferredLanguages(self.request)
        return upl.getSelectedLanguage()

    def languages(self):
        upl = IUserPreferredLanguages(self.request)
        if ICookieLanguageSelector.providedBy(upl):
            return upl.getLanguageList()
        return None


class INavigationMenuItemDirective(IViewletDirective):
    """An interface that makes title attribute of a viewlet translatable."""

    title = MessageID(
        title=u"The title for the menu item",
        required=False
        )

# Arbitrary keys and values are allowed to be passed to the viewlet.
INavigationMenuItemDirective.setTaggedValue('keyword_arguments', True)


class IDisableViewlet(Interface):
    pass


class DisableNavigationViewlet(BrowserView):
    implements(IDisableViewlet)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool flourish pages.
"""
import os
import re
import urllib

from zope.cachedescriptors.property import Lazy
from zope.component import getMultiAdapter, queryMultiAdapter
from zope.interface import implements, Interface
from zope.publisher.browser import BrowserPage
from zope.publisher.interfaces import NotFound
from zope.browser.interfaces import IBrowserView
from zope.traversing.api import getParent
from zope.traversing.browser.interfaces import IAbsoluteURL
from zope.traversing.browser.absoluteurl import absoluteURL, AbsoluteURL

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IApplicationTabs
from schooltool.skin.flourish.viewlet import Viewlet, ViewletManager
from schooltool.skin.flourish.viewlet import ManagerViewlet
from schooltool.skin.flourish import interfaces
from schooltool.skin.flourish import templates
from schooltool.traverser.traverser import PluggableTraverser, TraverserPlugin


class PageBase(BrowserPage):
    implements(interfaces.IPageBase)

    default_content_type = 'html'
    template = None

    def update(self):
        pass

    def render(self, *args, **kw):
        return self.template(*args, **kw)

    def publishTraverse(self, request, name):
        traverser = PluggableTraverser(self, request)
        return traverser.publishTraverse(request, name)

    @Lazy
    def providers(self):
        providers = getMultiAdapter(
            (self.context, self.request, self),
            interfaces.IContentProviders)
        return providers

    def __call__(self, *args, **kw):
        self.update()
        result = self.render(*args, **kw)
        return result


class Page(PageBase):
    implements(interfaces.IPage)

    title = None
    subtitle = None
    has_header = True
    page_class = 'page'
    container_class = 'container'

    template = templates.File('templates/main.pt')
    page_template = templates.File('templates/page.pt')
    content_template = None

    render = PageBase.render

    disabled = False

    def __call__(self, *args, **kw):
        disabled = self.disabled
        if (disabled and
            (not isinstance(disabled, (str, unicode)) or
             disabled.strip().lower() not in ('no', 'false'))):
            raise NotFound(self, self.__name__, self.request)
        self.update()
        if self.request.response.getStatus() in [300, 301, 302, 303,
                                                 304, 305, 307]:
            return u''
        result = self.render(*args, **kw)
        return result


class PageContentTraverser(TraverserPlugin):

    def __init__(self, view, request):
        self.view = view
        self.context = view.context
        self.request = request

    def traverse(self, name):
        parts = queryMultiAdapter(
            (self.context, self.request, self.view),
            interfaces.IContentProvider, name)
        if parts is None:
            raise NotFound(self.view, name, self.request)
        return parts


class PageAbsoluteURL(AbsoluteURL):

    def __str__(self):
        view = self.context
        request = self.request

        url = str(getMultiAdapter((view.context, request),
                                                 IAbsoluteURL))
        name = getattr(view, '__name__', None)
        if name:
            url += '/@@' + urllib.quote(name.encode('utf-8'), '@+')
        return url


class NoSidebarPage(Page):
    container_class = 'container extra-wide-container'

    page_template = templates.File('templates/page_nosidebar.pt')


class WideContainerPage(Page):
    container_class = 'container widecontainer'


class ContentViewletManager(ViewletManager):
    template = templates.Inline("""
        <tal:block repeat="viewlet view/viewlets">
          <div class="content"
               tal:define="rendered viewlet;
                           stripped rendered/strip|nothing"
               tal:condition="stripped"
               tal:content="structure stripped">
          </div>
        </tal:block>
    """)

    def render(self, *args, **kw):
        if not self.viewlets:
            return ''
        result = ViewletManager.render(self, *args, **kw)
        if result is not None:
            return result.strip()


class DisabledViewlet(Viewlet):

    def render(self, *args, **kw):
        return ''


class Refine(Viewlet):

    template = templates.Inline('''
      <div class="header"
           tal:condition="view/title"
           tal:content="view/title">
        [ Filter title ]
      </div>
      <div class="body" tal:content="structure view/body_template">
        [ options ]
      </div>
    ''')
    body_template = None
    render = lambda self, *a, **kw: self.template(*a, **kw)
    title = None


class Content(Viewlet):
    template = templates.File('templates/page_content.pt')
    body_template = None
    render = lambda self, *a, **kw: self.template(*a, **kw)


class Related(Viewlet):

    template = templates.Inline('''
      <div class="header"
           tal:condition="view/title"
           tal:content="view/title">
        [ Title ]
      </div>
      <div class="body" tal:content="structure view/body_template">
        [ Related info ]
      </div>
    ''')

    body_template = None
    render = lambda self, *a, **kw: self.template(*a, **kw)
    title = None


def getViewletViewName(context, request, view, manager):
    return getattr(view, '__name__', None)


def getViewParentActiveViewletName(context, request, view, manager):
    parent = getattr(view, 'view', None)
    if parent is None:
        return None
    name = queryMultiAdapter(
        (context, request, parent, manager),
        interfaces.IActiveViewletName,
        name='',
        default=None)
    return name


def getParentActiveViewletName(context, request, view, manager):
    parent = getParent(context)
    if parent is None:
        return None
    name = queryMultiAdapter(
        (parent, request, view, manager),
        interfaces.IActiveViewletName,
        name='',
        default=None)
    return name


class ListNavigationBase(object):
    template = templates.Inline("""
        <ul tal:attributes="class view/list_class"
            tal:condition="view/items">
          <li tal:repeat="item view/items"
              tal:attributes="class item/class"
              tal:content="structure item/viewlet">
          </li>
        </ul>
    """)
    list_class = None

    @Lazy
    def all_items(self):
        result = []
        active = self.active_viewlet
        for viewlet in self.viewlets:
            html_classes = []
            if viewlet.__name__ == active:
                html_classes.append('active')
            viewlet_class = getattr(viewlet, 'html_class', None)
            if viewlet_class:
                html_classes.append(viewlet_class)
            result.append({
                'class': ' '.join(html_classes) or None,
                'viewlet': viewlet,
                'content': viewlet(),
                })
        return result

    @Lazy
    def items(self):
        result = [item for item in self.all_items
                  if item['content'] and item['content'].strip()]
        return result

    @Lazy
    def active_viewlet(self):
        view = self.view
        name = queryMultiAdapter(
            (view.context, view.request, view, self),
            interfaces.IActiveViewletName,
            name='',
            default=None)
        return name

    @property
    def active(self):
        name = self.active_viewlet
        return(name and name in self.order)


class ListNavigationContent(ListNavigationBase, ViewletManager):
    pass


class ListNavigationViewlet(ListNavigationBase, ManagerViewlet):
    pass


class HeaderNavigationManager(ListNavigationContent):
    template = templates.Inline("""
        <ul tal:attributes="class view/list_class">
          <li tal:repeat="item view/items"
              tal:attributes="class item/class"
              tal:content="structure item/viewlet">
          </li>
        </ul>
    """)
    list_class = "navigation"

    @Lazy
    def viewlets(self):
        if self.cache is None:
            self.collect()
        app = ISchoolToolApplication(None)
        apptabs = IApplicationTabs(app)
        return [self[key] for key in self.order if apptabs.get(key, True)]


class SecondaryNavigationManager(ListNavigationContent):
    list_class = "second-nav"


class TertiaryNavigationManager(ListNavigationContent):
    template = templates.Inline("""
        <ul tal:attributes="class view/list_class"
            tal:condition="view/items">
          <li tal:repeat="item view/items"
              tal:attributes="class item/class"
              tal:content="structure item/viewlet">
          </li>
        </ul>
    """)

    @property
    def list_class(self):
        related_manager = queryMultiAdapter(
            (self.context, self.request, self.view),
            interfaces.IContentProvider,
            'page_related')
        # XXX: this check of viewlets might not be safe
        if related_manager is not None and related_manager.viewlets:
            return 'third-nav third-nav-narrow'
        return 'third-nav'


class PageNavigationManager(ListNavigationContent):
    list_class = "navigation"

    def render(self, *args, **kw):
        if not self.active:
            return ''
        return ViewletManager.render(self, *args, **kw)


class RefineLinksViewlet(Refine, ListNavigationViewlet):
    list_class = "filter"
    body_template = templates.Inherit(ListNavigationBase.template)

    def render(self, *args, **kw):
        if not self.items:
            return ''
        return Refine.render(self, *args, **kw)


class IHTMLHeadManager(interfaces.IViewletManager):
    pass


class IHeaderNavigationManager(interfaces.IViewletManager):
    pass


class ISecondaryNavigationManager(interfaces.IViewletManager):
    pass


class ITertiaryNavigationManager(interfaces.IViewletManager):
    pass


class IPageNavigationManager(interfaces.IViewletManager):
    pass


class IPageRefineManager(interfaces.IViewletManager):
    pass


class IPageContentManager(interfaces.IViewletManager):
    pass


class IContentActionsManager(interfaces.IViewletManager):
    pass


class IPageRelatedManager(interfaces.IViewletManager):
    pass


class LinkViewlet(Viewlet):
    template = templates.Inline('''
    <tal:block define="url view/url">
      <a tal:condition="url"
         tal:attributes="href view/url;
                         class view/css_class"
         tal:content="view/title"></a>
      <span tal:condition="not:url"
            tal:content="view/title"></span>
    </tal:block>
    ''')

    title = None
    css_class = None

    @property
    def enabled(self):
        return bool(self.title)

    @property
    def link(self):
        if not self.__name__:
            return None
        return self.__name__

    @property
    def url(self):
        link = self.link
        if not link:
            return None
        return "%s/%s" % (absoluteURL(self.context, self.request),
                          self.link)

    def render(self, *args, **kw):
        if not self.enabled:
            return ''
        return self.template(*args, **kw)


_invalid_html_id_characters = re.compile('[^a-zA-Z0-9-._]')


def sanitize_id(html_id):
    if not html_id:
        return html_id
    html_id = html_id.encode('punycode')
    html_id = _invalid_html_id_characters.sub(
        lambda m: ':'+str(ord(m.group(0))), html_id)
    if not html_id[0].isalpha():
        html_id = 'i' + html_id
    return html_id


def obj_random_html_id(obj, prefix='', len=6):
    name_list = ([
            'o'+os.urandom(len).encode('hex'),
            getattr(obj, '__name__', ''),
            prefix])
    return sanitize_id('-'.join(reversed(filter(None, name_list))))


def generic_viewlet_html_id(viewlet, prefix=''):
    parent = viewlet.manager.__parent__
    parents = []
    while (parent is not None and
           IBrowserView.providedBy(parent)):
        name = getattr(parent, '__name__', None)
        parents.append(str(name))
        parent = parent.__parent__
    name_list = ([str(viewlet.__name__),
                  getattr(viewlet.manager, '__name__', 'manager')] +
                 parents[:-1] +
                 [prefix])
    return sanitize_id('-'.join(reversed(name_list)))


class LinkIdViewlet(LinkViewlet):
    template = templates.Inline('''
    <tal:block define="url view/url">
      <a tal:condition="url"
         tal:attributes="href view/url;
                         id view/html_id"
         tal:content="view/title"></a>
      <span tal:condition="not:url"
            tal:attributes="id view/html_id"
            tal:content="view/title"></span>
    </tal:block>
    ''')

    @property
    def html_id(self):
        return generic_viewlet_html_id(self, 'LinkIdViewlet')


class SimpleModalLinkViewlet(LinkIdViewlet):
    template = templates.Inline('''
    <tal:block define="url view/url">
      <a tal:condition="url"
         href="#"
         tal:attributes="id view/html_id;
                         onclick view/onclick"
         tal:content="view/title"></a>
      <span tal:condition="not:url"
            tal:attributes="id view/html_id"
            tal:content="view/title"></span>
    </tal:block>
    ''')

    @property
    def onclick(self):
        return "return ST.dialogs.open_modal_form('%s', '%s');" % (
            self.url, self.html_id+'-container')


class ModalFormLinkViewlet(LinkIdViewlet):
    template = templates.File('templates/modal_form_link.pt')

    @property
    def form_container_id(self):
        return '%s-container' % self.html_id



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Upgrade SchoolTool to generation 31.

Contact relationships are expected to have extra_info now.

Also, fix relationship type: change URIContact to URIContactRelationship.
"""
from zope.app.generations.utility import findObjectsProviding
from zope.app.publication.zopepublication import ZopePublication

from schooltool.basicperson.interfaces import IBasicPerson
from schooltool.relationship.interfaces import IRelationshipLinks
from schooltool.relationship.relationship import relate, unrelate
from schooltool.contact.contact import URIContactRelationship
from schooltool.contact.contact import URIContact, URIPerson
from schooltool.contact.contact import ContactPersonInfo


def evolve(context):
    root = context.connection.root().get(ZopePublication.root_name, None)

    persons = findObjectsProviding(root, IBasicPerson)
    for person in persons:
        linkset = IRelationshipLinks(person)
        links = linkset.getLinksByRole(URIContact)
        for link in links[:]:
            if link.rel_type != URIContact:
                continue
            contact = link.target

            unrelate(URIContact,
                     (person, URIPerson),
                     (contact, URIContact))

            info = ContactPersonInfo()
            info.__parent__ = person
            relate(URIContactRelationship,
                   (person, URIPerson),
                   (contact, URIContact),
                   extra_info=info)




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Group objects
"""
__docformat__ = 'restructuredtext'
from persistent import Persistent

from zope.annotation.interfaces import IAttributeAnnotatable
from zope.container.contained import Contained
from zope.container.btree import BTreeContainer
from zope.lifecycleevent.interfaces import IObjectAddedEvent
from zope.lifecycleevent.interfaces import IObjectRemovedEvent
from zope.lifecycleevent import ObjectAddedEvent
from zope.app.dependable.interfaces import IDependable
from zope.intid import addIntIdSubscriber
from zope.intid.interfaces import IIntIds
from zope.component import adapter
from zope.component import adapts
from zope.component import getUtility
from zope.interface import implementer
from zope.interface import implements

from schooltool.app.app import Asset
from schooltool.app.app import InitBase
from schooltool.app.interfaces import ICalendarParentCrowd
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.membership import GroupMemberCrowd
from schooltool.app.membership import URIMembership, URIMember, URIGroup
from schooltool.app.security import ConfigurableCrowd
from schooltool.app.security import LeaderCrowd
from schooltool.course.interfaces import ISection
from schooltool.group import interfaces
from schooltool.group.interfaces import IGroupContainer
from schooltool.relationship import RelationshipProperty
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.schoolyear.subscriber import ObjectEventAdapterSubscriber
from schooltool.securitypolicy.crowds import AggregateCrowd
from schooltool.securitypolicy.crowds import AdministratorsCrowd
from schooltool.securitypolicy.crowds import ManagersCrowd
from schooltool.securitypolicy.crowds import ClerksCrowd
from schooltool.common import SchoolToolMessage as _


defaultGroups =  {"manager"       : _("Site Managers"),
                  "students"      : _("Students"),
                  "teachers"      : _("Teachers"),
                  "clerks"        : _("Clerks"),
                  "administrators": _("School Administrators"),
                  }


defaultManagerGroups = ("manager", "clerks")


class GroupContainerContainer(BTreeContainer):
    """Container of group containers."""

    implements(interfaces.IGroupContainerContainer,
               IAttributeAnnotatable)


class GroupContainer(BTreeContainer):
    """Container of groups."""

    implements(interfaces.IGroupContainer, IAttributeAnnotatable)


@adapter(ISchoolYear)
@implementer(interfaces.IGroupContainer)
def getGroupContainer(sy):
    addIntIdSubscriber(sy, ObjectAddedEvent(sy))
    int_ids = getUtility(IIntIds)
    sy_id = str(int_ids.getId(sy))
    app = ISchoolToolApplication(None)
    gc = app['schooltool.group'].get(sy_id, None)
    if gc is None:
        gc = app['schooltool.group'][sy_id] = GroupContainer()
    return gc


@adapter(ISchoolToolApplication)
@implementer(interfaces.IGroupContainer)
def getGroupContainerForApp(app):
    syc = ISchoolYearContainer(app)
    sy = syc.getActiveSchoolYear()
    if sy is None:
        return None
    return IGroupContainer(sy)


@adapter(ISection)
@implementer(interfaces.IGroupContainer)
def getGroupContainerForSection(section):
    sy = ISchoolYear(section)
    return IGroupContainer(sy)


@adapter(interfaces.IGroupContainer)
@implementer(ISchoolYear)
def getSchoolYearForGroupContainer(group_container):
    container_id = int(group_container.__name__)
    int_ids = getUtility(IIntIds)
    container = int_ids.getObject(container_id)
    return container


class Group(Persistent, Contained, Asset):
    """Group."""

    implements(interfaces.IGroup, interfaces.IGroupContained,
               IAttributeAnnotatable)

    members = RelationshipProperty(URIMembership, URIGroup, URIMember)

    def __init__(self, title=None, description=None):
        self.title = title
        self.description = description


# XXX Unit test ME!
class InitGroupsForNewSchoolYear(ObjectEventAdapterSubscriber):

    adapts(IObjectAddedEvent, ISchoolYear)

    def initializeGroupContainer(self):
        groups = IGroupContainer(self.object)
        for id, title in defaultGroups.items():
            group = groups[id] = Group(title)
            IDependable(group).addDependent('')
        persons = ISchoolToolApplication(None)['persons']
        manager = persons.super_user
        if manager is None:
            return
        for id in defaultManagerGroups:
            if manager not in groups[id].members:
                groups[id].members.add(manager)

    def importDefaultGroups(self, activeSchoolyear):
        oldGroups = IGroupContainer(activeSchoolyear)
        newGroups = IGroupContainer(self.object)
        for groupId in defaultGroups:
            if groupId in oldGroups:
                oldGroup = oldGroups[groupId]
                newGroup = Group(oldGroup.title, oldGroup.description)
                newGroups[groupId] = newGroup
                IDependable(newGroup).addDependent('')

    def __call__(self):
        app = ISchoolToolApplication(None)
        syc = ISchoolYearContainer(app)
        activeSchoolyear = syc.getActiveSchoolYear()
        if activeSchoolyear is not None:
            self.importDefaultGroups(activeSchoolyear)
        else:
            self.initializeGroupContainer()


class GroupInit(InitBase):

    def __call__(self):
        self.app['schooltool.group'] = GroupContainerContainer()


class GroupContainerViewersCrowd(ConfigurableCrowd):
    setting_key = 'everyone_can_view_group_list'


class GroupViewersCrowd(ConfigurableCrowd):
    setting_key = 'everyone_can_view_group_info'


class GroupCalendarSettingCrowd(ConfigurableCrowd):
    setting_key = 'everyone_can_view_group_calendar'


class GroupCalendarViewersCrowd(AggregateCrowd):
    implements(ICalendarParentCrowd)
    adapts(interfaces.IGroup)

    def crowdFactories(self):
        return [GroupCalendarSettingCrowd, GroupMemberCrowd, LeaderCrowd,
                AdministratorsCrowd, ManagersCrowd, ClerksCrowd]


class GroupCalendarMemberSettingCrowd(ConfigurableCrowd):
    setting_key = 'members_can_edit_group_calendar'

    def contains(self, principal):
        return (ConfigurableCrowd.contains(self, principal) and
                GroupMemberCrowd(self.context).contains(principal))


class GroupCalendarEditorsCrowd(AggregateCrowd):
    implements(ICalendarParentCrowd)
    adapts(interfaces.IGroup)

    def crowdFactories(self):
        return [GroupCalendarMemberSettingCrowd, LeaderCrowd]


class RemoveGroupsWhenSchoolYearIsDeleted(ObjectEventAdapterSubscriber):
    adapts(IObjectRemovedEvent, ISchoolYear)

    def __call__(self):
        group_container = IGroupContainer(self.object)
        for group_id, group in list(group_container.items()):
            IDependable(group).removeDependent('')
            del group_container[group_id]



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Views for school years and school year container implementation
"""
import cPickle
from StringIO import StringIO

from zope.viewlet.viewlet import ViewletBase
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.cachedescriptors.property import Lazy
from zope.component import adapts, getMultiAdapter
from zope.event import notify
from zope.security import checkPermission
from zope.schema import Date, TextLine
from zope.schema.interfaces import ValidationError
from zope.interface.exceptions import Invalid
from zope.interface import implements
from zope.interface import Interface
from zope.location.pickling import CopyPersistent
from zope.traversing.browser.absoluteurl import AbsoluteURL
from zope.traversing.browser.interfaces import IAbsoluteURL
from zope.traversing.browser import absoluteURL
from zope.container.interfaces import INameChooser
from zope.container.contained import containedEvent
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.proxy import sameProxiedObjects
from zope.i18n import translate
from zope.i18n.interfaces.locales import ICollator
from zope.security.checker import canAccess

from z3c.form import form, field, button
from z3c.form.util import getSpecification
from z3c.form.validator import NoInputData
from z3c.form.validator import WidgetsValidatorDiscriminators
from z3c.form.validator import InvariantsValidator
from z3c.form.validator import WidgetValidatorDiscriminators
from z3c.form.validator import SimpleFieldValidator
from z3c.form.error import ErrorViewSnippet
from z3c.form.interfaces import DISPLAY_MODE

import zc.table.column

import schooltool.skin.flourish.containers
import schooltool.skin.flourish.breadcrumbs
from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.common import DateRange
from schooltool.common import SchoolToolMessage as _
from schooltool.common.inlinept import InheritTemplate
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.course.interfaces import ICourseContainer
from schooltool.course.course import Course
from schooltool.group.interfaces import IGroupContainer
from schooltool.group.group import Group, defaultGroups
from schooltool.schoolyear.browser.interfaces import ISchoolYearViewMenuViewletManager
from schooltool.schoolyear.schoolyear import validateScholYearForOverflow
from schooltool.schoolyear.schoolyear import validateScholYearsForOverlap
from schooltool.schoolyear.schoolyear import SchoolYear
from schooltool.schoolyear.interfaces import TermOverflowError
from schooltool.schoolyear.interfaces import SchoolYearOverlapError
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.skin.skin import OrderedViewletManager
from schooltool.skin.containers import ContainerDeleteView
from schooltool.skin.containers import TableContainerView
from schooltool.skin import flourish
from schooltool.table import table
from schooltool.timetable.interfaces import ITimetableContainer
from schooltool.timetable.timetable import Timetable
from schooltool.timetable.interfaces import IWeekDayTemplates
from schooltool.timetable.interfaces import ISchoolDayTemplates
from schooltool.timetable.daytemplates import WeekDayTemplates
from schooltool.timetable.daytemplates import SchoolDayTemplates
from schooltool.timetable.daytemplates import DayTemplate
from schooltool.timetable.daytemplates import TimeSlot
from schooltool.timetable.schedule import Period

class SchoolYearContainerAbsoluteURLAdapter(AbsoluteURL):

    adapts(ISchoolYearContainer, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        app = ISchoolToolApplication(None)
        url = str(getMultiAdapter((app, self.request), name='absolute_url'))
        return url + '/schoolyears'

    __call__ = __str__


class SchoolYearAbsoluteURLAdapter(AbsoluteURL):

    adapts(ISchoolYear, IBrowserRequest)
    implements(IAbsoluteURL)


class SchoolYearContainerBaseView(object):

    def canDeleteSchoolYears(self):
        return len(self.listIdsForDeletion()) >= len(self.context)

    def deletingActiveSchoolYear(self):
        return self.context.getActiveSchoolYear().__name__ in self.listIdsForDeletion()


class SchoolYearContainerDeleteView(ContainerDeleteView, SchoolYearContainerBaseView):
    """A view for deleting items from container."""

    def update(self):
        if 'CONFIRM' in self.request:
            for key in self.listIdsForDeletion():
                if key != self.context.active_id:
                    del self.context[key]
            if self.deletingActiveSchoolYear():
                del self.context[self.context.active_id]
            self.request.response.redirect(self.nextURL())
        elif 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())


class SchoolYearContainerView(TableContainerView, SchoolYearContainerBaseView):
    """SchoolYear container view."""

    __used_for__ = ISchoolYearContainer
    template = ViewPageTemplateFile("templates/schoolyear_container.pt")

    index_title = _("School Years")
    error = None

    @property
    def first(self):
        year = self.context.getActiveSchoolYear()
        if year is not None:
            return year.first

    @property
    def last(self):
        year = self.context.getActiveSchoolYear()
        if year is not None:
            return year.last

    def setUpTableFormatter(self, formatter):
        columns_before = []
        if self.canModify():
            columns_before = [
                table.DependableCheckboxColumn(prefix="delete",
                                               name='delete_checkbox',
                                               title=u'')]
        columns_after = [table.DateColumn(title=_("Starts"),
                                          getter=lambda x, y: x.first),
                         table.DateColumn(title=_("Ends"),
                                          getter=lambda x, y: x.last)]
        formatter.setUp(formatters=[table.url_cell_formatter],
                        columns_before=columns_before,
                        columns_after=columns_after)

    def __call__(self):
        if 'DELETE' in self.request:
            if not self.deletingActiveSchoolYear():
                return self.delete_template()
            elif not self.canDeleteSchoolYears():
                self.error = _("You can not delete the active school year."
                               " Unless you are deleting all the school years.")
            else:
                return self.delete_template()

        self.setUpTableFormatter(self.table)
        return self.template()

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    def update(self):
        if 'ACTIVATE_NEXT_SCHOOLYEAR' in self.request:
            for key in self.listIdsForDeletion():
                if key in self.context:
                    self.context.activateNextSchoolYear(key)
                    self.request.response.redirect(self.nextURL())


class FlourishActiveSchoolYearColumn(zc.table.column.Column):
    """Table column that displays whether a schoolyear is the active one.
    """

    def renderCell(self, item, formatter):
        if item.__parent__.active_id == item.__name__:
            return '<span class="ui-icon ui-icon-check"></span>'
        else:
            return ''


class SchoolYearTableFormatter(table.SchoolToolTableFormatter):

    def sortOn(self):
        return (('first', True),)


class FlourishSchoolYearContainerView(table.TableContainerView):
    """flourish SchoolYear container view."""

    def getColumnsAfter(self):
        result = [
            table.DateColumn(title=_("First Day"),
                             name='first',
                             getter=lambda x, y: x.first),
            table.DateColumn(title=_("Last Day"),
                             getter=lambda x, y: x.last),
            FlourishActiveSchoolYearColumn(title=_("Active")),
            ]
        return result


class FlourishSchoolYearContainerLinks(flourish.page.RefineLinksViewlet):
    """SchoolYear container links viewlet."""


class FlourishSchoolYearContainerActionLinks(flourish.page.RefineLinksViewlet):
    """SchoolYear container action links viewlet."""


class FlourishSchoolYearActionLinks(flourish.page.RefineLinksViewlet):
    """SchoolYear action links viewlet."""


class FlourishSchoolYearDeleteLink(flourish.page.ModalFormLinkViewlet):

    @property
    def dialog_title(self):
        title = _(u'Delete ${schoolyear}',
                  mapping={'schoolyear': self.context.title})
        return translate(title, context=self.request)


class FlourishSchoolYearDeleteView(flourish.form.DialogForm, form.EditForm):
    """View used for confirming deletion of a schoolyear."""

    dialog_submit_actions = ('apply',)
    dialog_close_actions = ('cancel',)
    label = None

    @button.buttonAndHandler(_("Delete"), name='apply')
    def handleDelete(self, action):
        url = '%s/delete.html?delete.%s&CONFIRM' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.__name__.encode('utf-8'))
        self.request.response.redirect(url)
        # We never have errors, so just close the dialog.
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        pass

    def updateActions(self):
        super(FlourishSchoolYearDeleteView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class ISchoolYearAddForm(Interface):

    title = TextLine(
        title=_("Title"))

    first = Date(
        title=_(u"First day"))

    last = Date(
        title=_(u"Last day"))


class SchoolYearAddFormAdapter(object):
    implements(ISchoolYearAddForm)
    adapts(ISchoolYear)

    def __init__(self, context):
        self.__dict__['context'] = context

    def __setattr__(self, name, value):
        setattr(self.context, name, value)

    def __getattr__(self, name):
        return getattr(self.context, name)


class ImportSchoolYearData(object):

    def hasCourses(self, schoolyear):
        courses = ICourseContainer(schoolyear)
        return bool(courses)

    def hasTimetableSchemas(self, schoolyear):
        timetables = ITimetableContainer(schoolyear)
        return bool(timetables)

    def activeSchoolyearInfo(self):
        result = {}
        request = self.request
        collator = ICollator(request.locale)
        activeSchoolyear = self.context.getActiveSchoolYear()
        if activeSchoolyear is not None:
            result['title'] = activeSchoolyear.title
            result['hasCourses'] = self.hasCourses(activeSchoolyear)
            result['hasTimetables'] = self.hasTimetableSchemas(activeSchoolyear)
            result['groups'] = []
            groups = IGroupContainer(activeSchoolyear)
            for groupId, group in sorted(groups.items(),
                                         cmp=collator.cmp,
                                         key=lambda (groupId,group):group.title):
                info = {}
                info['id'] = groupId
                info['title'] = group.title
                info['isDefault'] = groupId in defaultGroups
                info['hasMembers'] = bool(list(group.members))
                info['sent'] = groupId in self.customGroupsToImport
                info['membersSent'] = groupId in self.groupsWithMembersToImport
                result['groups'].append(info)
        return result

    def importAllCourses(self):
        if not self.shouldImportAllCourses():
            return
        oldCourses = ICourseContainer(self.activeSchoolyear)
        newCourses = ICourseContainer(self.newSchoolyear)
        for id, course in oldCourses.items():
            newCourses[course.__name__] = new_course = Course(course.title, course.description)
            new_course.course_id = course.course_id
            new_course.government_id = course.government_id
            new_course.credits = course.credits

    def setUpTimetable(self, timetable, old_timetable):
        if IWeekDayTemplates.providedBy(old_timetable.time_slots):
            timetable.time_slots, object_event = containedEvent(
                WeekDayTemplates(), timetable, 'time_slots')
        elif ISchoolDayTemplates.providedBy(old_timetable.time_slots):
            timetable.time_slots, object_event = containedEvent(
                SchoolDayTemplates(), timetable, 'time_slots')
        notify(object_event)
        timetable.time_slots.initTemplates()
        old_templates = old_timetable.time_slots.templates
        for old_template_key, old_template in old_templates.items():
            template = DayTemplate(old_template.title)
            timetable.time_slots.templates[old_template_key] = template
            for old_timeslot_key, old_timeslot in old_template.items():
                timeslot = TimeSlot(old_timeslot.tstart,
                                    old_timeslot.duration,
                                    old_timeslot.activity_type)
                template[old_timeslot_key] = timeslot
        if IWeekDayTemplates.providedBy(old_timetable.periods):
            timetable.periods, object_event = containedEvent(
                WeekDayTemplates(), timetable, 'periods')
        elif ISchoolDayTemplates.providedBy(old_timetable.periods):
            timetable.periods, object_event = containedEvent(
                SchoolDayTemplates(), timetable, 'periods')
        notify(object_event)
        timetable.periods.initTemplates()
        old_templates = old_timetable.periods.templates
        for old_template_key, old_template in old_templates.items():
            template = DayTemplate(old_template.title)
            timetable.periods.templates[old_template_key] = template
            for old_period_key, old_period in old_template.items():
                period = Period(old_period.title, old_period.activity_type)
                template[old_period_key] = period

    def importAllTimetables(self):
        if not self.shouldImportAllTimetables():
            return
        oldTimetables = ITimetableContainer(self.activeSchoolyear)
        newTimetables = ITimetableContainer(self.newSchoolyear)
        chooser = INameChooser(newTimetables)
        app = ISchoolToolApplication(None)
        tzname = IApplicationPreferences(app).timezone
        for schooltt in oldTimetables.values():
            newSchooltt = Timetable(
                self.newSchoolyear.first, self.newSchoolyear.last,
                title=schooltt.title,
                timezone=tzname)
            name = chooser.chooseName(schooltt.__name__, newSchooltt)
            newTimetables[name] = newSchooltt
            self.setUpTimetable(newSchooltt, schooltt)
            if (oldTimetables.default is not None and
                sameProxiedObjects(oldTimetables.default, schooltt)):
                newTimetables.default = newSchooltt

    def importGroupMembers(self, sourceGroup, targetGroup):
        for member in sourceGroup.members:
            targetGroup.members.add(member)

    def importGroup(self, groupId, shouldImportMembers=False):
        oldGroups = IGroupContainer(self.activeSchoolyear)
        newGroups = IGroupContainer(self.newSchoolyear)
        if groupId in oldGroups:
            oldGroup = oldGroups[groupId]
            newGroup = newGroups[groupId] = Group(oldGroup.title,
                                                   oldGroup.description)
            if shouldImportMembers:
                self.importGroupMembers(oldGroup, newGroup)

    def importCustomGroups(self):
        for groupId in self.customGroupsToImport:
            shouldImportMembers = groupId in self.groupsWithMembersToImport
            self.importGroup(groupId, shouldImportMembers)

    def importDefaultGroupsMembers(self):
        oldGroups = IGroupContainer(self.activeSchoolyear)
        newGroups = IGroupContainer(self.newSchoolyear)
        for groupId in defaultGroups:
            if groupId in oldGroups and groupId in self.groupsWithMembersToImport:
                oldGroup = oldGroups[groupId]
                newGroup = newGroups[groupId]
                self.importGroupMembers(oldGroup, newGroup)

    def shouldImportData(self):
        return self.activeSchoolyear is not None and \
               not sameProxiedObjects(self.activeSchoolyear, self.newSchoolyear)

    def shouldImportAllCourses(self):
        return 'importAllCourses' in self.request and \
               self.hasCourses(self.activeSchoolyear)

    def shouldImportAllTimetables(self):
        return 'importAllTimetables' in self.request and \
               self.hasTimetableSchemas(self.activeSchoolyear)

    @property
    def customGroupsToImport(self):
        result = self.request.get('groups', [])
        if not isinstance(result, list):
            result = [result]
        return result

    @property
    def groupsWithMembersToImport(self):
        result = self.request.get('members', [])
        if not isinstance(result, list):
            result = [result]
        return result

    def importData(self, newSchoolyear):
        self.newSchoolyear = newSchoolyear
        self.activeSchoolyear = self.context.getActiveSchoolYear()
        if self.shouldImportData():
            self.importAllCourses()
            self.importAllTimetables()
            self.importCustomGroups()
            self.importDefaultGroupsMembers()


class SchoolYearAddView(form.AddForm, ImportSchoolYearData):
    """School Year add form for school years."""
    label = _("Add new school year")
    template = ViewPageTemplateFile('templates/schoolyear_add.pt')

    fields = field.Fields(ISchoolYearAddForm)

    def updateActions(self):
        super(SchoolYearAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(_('Add'), name='add')
    def handleAdd(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return

        obj = self.createAndAdd(data)

        if obj is not None:
            # mark only as finished if we get the new object
            self._finishedAdd = True

    def create(self, data):
        schoolyear = SchoolYear(data['title'], data['first'], data['last'])
        form.applyChanges(self, schoolyear, data)
        self._schoolyear = schoolyear
        return schoolyear

    def nextURL(self):
        return absoluteURL(self._schoolyear, self.request)

    def add(self, schoolyear):
        """Add `schoolyear` to the container."""
        chooser = INameChooser(self.context)
        name = chooser.chooseName(schoolyear.title, schoolyear)
        self.context[name] = schoolyear
        self.importData(schoolyear)
        return schoolyear

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)


class FlourishSchoolYearAddView(flourish.form.AddForm, SchoolYearAddView):

    template = InheritTemplate(flourish.page.Page.template)
    label = None
    legend = _('School Year Details')

    @button.buttonAndHandler(_('Submit'), name='add')
    def handleAdd(self, action):
        super(FlourishSchoolYearAddView, self).handleAdd.func(self, action)

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        super(FlourishSchoolYearAddView, self).handle_cancel_action.func(self,
            action)

    def updateWidgets(self):
        super(FlourishSchoolYearAddView, self).updateWidgets()
        self.widgets['title'].maxlength = 12
        title_description =  _('Limited to 12 characters or less')
        self.widgets['title'].field.description = title_description


class SchoolYearEditView(form.EditForm):
    """Edit form for basic person."""
    form.extends(form.EditForm)
    template = ViewPageTemplateFile('templates/schoolyear_edit.pt')

    fields = field.Fields(ISchoolYearAddForm)

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(SchoolYearEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @property
    def label(self):
        return _(u'Change information for ${schoolyear_title}',
                 mapping={'schoolyear_title': self.context.title})


class FlourishSchoolYearEditView(flourish.form.Form, SchoolYearEditView):
    """flourish Edit form for schoolyear."""

    template = InheritTemplate(flourish.page.Page.template)
    label = None

    def update(self):
        SchoolYearEditView.update(self)

    @button.buttonAndHandler(_('Submit'), name='apply')
    def handleApply(self, action):
        super(FlourishSchoolYearEditView, self).handleApply.func(self, action)
        if self.status == self.successMessage:
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateWidgets(self):
        super(FlourishSchoolYearEditView, self).updateWidgets()
        self.widgets['title'].maxlength = 12
        title_description =  _('Limited to 12 characters or less')
        self.widgets['title'].field.description = title_description

    @property
    def legend(self):
        return _(u'Change information for ${schoolyear_title}',
                 mapping={'schoolyear_title': self.context.title})


class AddSchoolYearOverlapValidator(InvariantsValidator):

    def validateObject(self, obj):
        errors = super(AddSchoolYearOverlapValidator, self).validateObject(obj)
        try:
            dr = DateRange(obj.first, obj.last)
            try:
                validateScholYearsForOverlap(self.view.context, dr, None)
            except SchoolYearOverlapError, e:
                errors += (e, )
        except ValueError, e:
            errors += (Invalid(_("School year must begin before it ends.")), )
        except NoInputData:
            return errors
        return errors

WidgetsValidatorDiscriminators(
    AddSchoolYearOverlapValidator,
    view=SchoolYearAddView,
    schema=getSpecification(ISchoolYearAddForm, force=True))


class EditSchoolYearValidator(InvariantsValidator):

    def validateObject(self, obj):
        errors = super(EditSchoolYearValidator, self).validateObject(obj)
        try:
            dr = DateRange(obj.first, obj.last)
            try:
                validateScholYearsForOverlap(self.view.context.__parent__, dr, self.view.context)
            except SchoolYearOverlapError, e:
                errors += (e, )

            try:
                validateScholYearForOverflow(dr, self.view.context)
            except TermOverflowError, e:
                errors += (e, )
        except ValueError, e:
            errors += (Invalid(_("School year must begin before it ends.")), )
        except NoInputData:
            return errors
        return errors


WidgetsValidatorDiscriminators(
    EditSchoolYearValidator,
    view=SchoolYearEditView,
    schema=getSpecification(ISchoolYearAddForm, force=True))


class FlourishInvalidDateRangeError(ValidationError):
    __doc__ = _('School year must begin before it ends')


class FlourishOverlapError(ValidationError):
    __doc__ = _('Date range overlaps another school year')


class FlourishOverflowError(ValidationError):
    __doc__ = _('Date range too small to contain the currently set up term(s)')


class FlourishOverlapValidator(SimpleFieldValidator):

    def validate(self, value):
        # XXX: hack to display the overlap error next to the widget!
        rv = super(FlourishOverlapValidator, self).validate(value)
        last_widget = self.view.widgets['last']
        last_value = self.request.get(last_widget.name)
        try:
            last_value = last_widget._toFieldValue(last_value)
        except:
            return
        try:
            dr = DateRange(value, last_value)
        except:
            raise FlourishInvalidDateRangeError()
        try:
            validateScholYearsForOverlap(self.container, dr, self.schoolyear)
        except SchoolYearOverlapError, e:
            raise FlourishOverlapError()
        if self.schoolyear:
            try:
                validateScholYearForOverflow(dr, self.schoolyear)
            except TermOverflowError, e:
                raise FlourishOverflowError()


class FlourishOverlapAddValidator(FlourishOverlapValidator):
    schoolyear = None

    @property
    def container(self):
        return self.context


class FlourishOverlapEditValidator(FlourishOverlapValidator):

    @property
    def schoolyear(self):
        return self.context

    @property
    def container(self):
        return self.context.__parent__


WidgetValidatorDiscriminators(FlourishOverlapAddValidator,
                              view=FlourishSchoolYearAddView,
                              field=ISchoolYearAddForm['first'])


WidgetValidatorDiscriminators(FlourishOverlapEditValidator,
                              view=FlourishSchoolYearEditView,
                              field=ISchoolYearAddForm['first'])


class OverlapErrorViewSnippet(ErrorViewSnippet):

    adapts(SchoolYearOverlapError, None, None, None, None, None)

    render = ViewPageTemplateFile("templates/school_year_overlap_error.pt")

    def schoolyears(self):
        return self.context.overlapping_schoolyears

    def createMessage(self):
        return self.context.__repr__()


class OverflowErrorViewSnippet(ErrorViewSnippet):

    adapts(TermOverflowError, None, None, None, None, None)

    render = ViewPageTemplateFile("templates/term_overflow_error.pt")

    def terms(self):
        return self.context.overflowing_terms

    def createMessage(self):
        return self.context.__repr__()


class SchoolYearView(TableContainerView):
    """School Year view."""

    __used_for__ = ISchoolYear
    delete_template = ViewPageTemplateFile("templates/term-delete.pt")
    template = ViewPageTemplateFile("templates/schoolyear.pt")

    index_title = _("School Year")

    @property
    def sorted_terms(self):
        return sorted(self.context.values(), key=lambda t: t.last)

    @property
    def first(self):
        return self.context.first

    @property
    def last(self):
        return self.context.last

    def update(self):
        if 'CONFIRM' in self.request:
            for key in self.listIdsForDeletion():
                del self.context[key]


class SchoolYearViewMenuViewletManager(OrderedViewletManager):
    """Viewlet manager for displaying the various menus at the top of a page."""

    implements(ISchoolYearViewMenuViewletManager)


class ActiveSchoolYears(ViewletBase):

    def activeSchoolYear(self):
        """Return the active school year."""
        return ISchoolYearContainer(ISchoolToolApplication(None)).getActiveSchoolYear()

    def nextSchoolYear(self):
        """Return the next school year."""
        syc = ISchoolYearContainer(ISchoolToolApplication(None))
        if checkPermission("schooltool.edit", syc):
            return syc.getNextSchoolYear()


class FlourishManageYearOverview(flourish.page.Content,
                                 ActiveSchoolYearContentMixin):

    body_template = ViewPageTemplateFile(
        'templates/f_manage_year_overview.pt')


class FlourishSchoolYearView(flourish.page.Page):
    """flourish SchoolYear view."""

    @property
    def subtitle(self):
        return self.context.title


class FlourishSchoolYearDetails(flourish.form.FormViewlet):

    fields = field.Fields(ISchoolYearAddForm).omit('title')
    template = ViewPageTemplateFile("templates/f_schoolyear.pt")
    mode = DISPLAY_MODE

    @property
    def canModify(self):
        return canAccess(self.context.__parent__, '__delitem__')


class FlourishSchoolYearActivateView(flourish.page.Page):

    message = None

    def years(self):
        for year in reversed(tuple(self.context.values())):
            yield {
                'obj': year,
                'active': year.__name__ == year.__parent__.active_id,
                }

    def update(self):
        if 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())
        if 'SUBMIT' in self.request:
            if self.request.get('ACTIVATE'):
                self.context.activateNextSchoolYear(self.request['ACTIVATE'])
                self.request.response.redirect(self.nextURL())
            else:
                self.message = _("Please select a school year before clicking "
                                 "'Submit'.")

    def nextURL(self):
        next = self.request.get('next')
        if next:
            app_url = absoluteURL(ISchoolToolApplication(None), self.request)
            return app_url + '/' + next
        return absoluteURL(self.context, self.request)


class ManageSchoolTertiaryNavigation(flourish.page.Content,
                                     flourish.page.TertiaryNavigationManager,
                                     ActiveSchoolYearContentMixin):

    template = InlineViewPageTemplate("""
        <ul tal:attributes="class view/list_class"
            tal:condition="view/items">
          <li tal:repeat="item view/items"
              tal:attributes="class item/class">
              <a tal:attributes="href item/url"
                 tal:content="item/schoolyear/@@title" />
          </li>
        </ul>
    """)

    @property
    def items(self):
        result = []
        active = self.schoolyear
        schoolyears = active.__parent__ if active is not None else {}
        for schoolyear in schoolyears.values():
            url = '%s/%s?schoolyear_id=%s' % (
                absoluteURL(self.context, self.request),
                'manage',
                schoolyear.__name__)
            css_class = schoolyear.first == active.first and 'active' or None
            result.append({
                    'class': css_class,
                    'url': url,
                    'schoolyear': schoolyear,
                    })
        return result


class SchoolyearNavBreadcrumbs(flourish.breadcrumbs.Breadcrumbs):

    traversal_name = u''

    @property
    def schoolyear_id(self):
        sy = ISchoolYear(self.context, None)
        if sy is None:
            return u''
        return sy.__name__

    @property
    def crumb_parent(self):
        return ISchoolToolApplication(None)

    @property
    def url(self):
        if not self.checkPermission():
            return False
        app = ISchoolToolApplication(None)
        app_url = absoluteURL(app, self.request)
        link = '%s/%s' % (app_url, self.traversal_name)
        sy_id = self.schoolyear_id
        if sy_id:
            link += '?schoolyear_id=%s' % sy_id
        return link


class FlourishSchoolYearsOverview(flourish.page.Content):

    body_template = ViewPageTemplateFile(
        'templates/f_manage_schoolyears_overview.pt')

    @Lazy
    def schoolyears(self):
        syc = ISchoolYearContainer(self.context)
        return syc


class SchoolYearAddLink(flourish.page.LinkViewlet):

    @property
    def url(self):
        schoolyears = ISchoolYearContainer(self.context)
        return absoluteURL(schoolyears, self.request) + '/add.html'


class FlourishActivateNewYearLink(flourish.page.LinkViewlet):

    @property
    def enabled(self):
        if not flourish.canEdit(self.schoolyears):
            return False
        return super(FlourishActivateNewYearLink, self).enabled

    @property
    def schoolyears(self):
        return ISchoolYearContainer(ISchoolToolApplication(None))

    @property
    def url(self):
        link = self.link
        if not link:
            return None
        return "%s/%s?next=%s" % (
            absoluteURL(self.schoolyears, self.request),
            self.link,
            self.view.__name__)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from zope.interface import implements, implementer
from zope.component import adapts, adapter
from zope.component import getAdapter, getMultiAdapter
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.publisher.browser import BrowserView
from zope.traversing.api import getName
from zope.traversing.browser.interfaces import IAbsoluteURL
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.container.interfaces import INameChooser
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.schema import ValidationError

from z3c.form.interfaces import ITextAreaWidget
from z3c.form import form, field, button, validator
from z3c.form.converter import BaseDataConverter, FormatterValidationError

import schooltool.skin.flourish.form
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.basicperson.demographics import TextFieldDescription
from schooltool.basicperson.demographics import DescriptionFieldDescription
from schooltool.basicperson.demographics import DateFieldDescription
from schooltool.basicperson.demographics import BoolFieldDescription
from schooltool.basicperson.demographics import EnumFieldDescription
from schooltool.basicperson.demographics import IntFieldDescription
from schooltool.basicperson.interfaces import IEnumFieldDescription
from schooltool.basicperson.interfaces import IIntFieldDescription
from schooltool.basicperson.interfaces import IDemographicsFields
from schooltool.basicperson.interfaces import IFieldDescription
from schooltool.basicperson.interfaces import EnumValueList
from schooltool.basicperson.interfaces import IAddEditViewTitle
from schooltool.basicperson.interfaces import ILimitKeysLabel
from schooltool.basicperson.interfaces import ILimitKeysHint
from schooltool.common.inlinept import InheritTemplate
from schooltool.skin import flourish

from schooltool.common import format_message
from schooltool.common import SchoolToolMessage as _


class CustomEnumDataConverter(BaseDataConverter):
    """A special data converter for iso enums."""

    adapts(EnumValueList, ITextAreaWidget)

    def toWidgetValue(self, value):
        """See interfaces.IDataConverter"""
        if value is self.field.missing_value:
            return u''
        return unicode('\n'.join(value))

    def toFieldValue(self, value):
        """See interfaces.IDataConverter"""
        if value == u'':
            return self.field.missing_value
        lines = filter(None, [s.strip() for s in unicode(value).splitlines()])
        if not lines:
            return self.field.missing_value
        field_value = []
        for line in lines:
            if line in field_value:
                raise FormatterValidationError(
                    format_message(_('Duplicate entry "${value}"'),
                                   mapping={'value': line}),
                    line)
            if len(line) >= 64:
                raise FormatterValidationError(
                    format_message(_('Value too long "${value}"'),
                                   mapping={'value': line}),
                    line)
            field_value.append(line)
        return field_value


class DemographicsFieldsAbsoluteURLAdapter(BrowserView):

    adapts(IDemographicsFields, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        app = ISchoolToolApplication(None)
        url = str(getMultiAdapter((app, self.request), name='absolute_url'))
        return url + '/demographics'

    __call__ = __str__


class DemographicsView(BrowserView):
    """A Demographics List view."""

    title = _('Demographics and Personal Information')

    def demographics(self):
        pos = 0
        for demo in list(self.context.values()):
            pos += 1
            yield {'name': demo.__name__,
                   'title': demo.name,
                   'url': absoluteURL(demo, self.request),
                   'pos': pos}

    def positions(self):
        return range(1, len(self.context.values())+1)

    def update(self):
        if 'DELETE' in self.request:
            for name in self.request.get('delete', []):
                del self.context[name]
        elif 'form-submitted' in self.request:
            old_pos, new_pos, move_detected = 0, 0, False
            for activity in self.context.values():
                old_pos += 1
                name = getName(activity)
                if 'pos.'+name not in self.request:
                    continue
                new_pos = int(self.request['pos.'+name])
                if new_pos != old_pos:
                    move_detected = True
                    break
            old_pos, new_pos = old_pos-1, new_pos-1
            keys = list(self.context.keys())
            moving = keys[old_pos]
            keys.remove(moving)
            keys.insert(new_pos,moving)
            self.context.updateOrder(keys)


class FieldDescriptionAddView(form.AddForm):
    """Add form for Field Descriptions."""

    label = _("Add new Field")
    form.extends(form.AddForm)
    template = ViewPageTemplateFile('templates/person_add.pt')
    fields = field.Fields(IFieldDescription)

    def updateActions(self):
        super(FieldDescriptionAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        return absoluteURL(self._fd, self.request)

    def add(self, fd):
        """Add field description to the container."""
        chooser = INameChooser(self.context)
        name = chooser.chooseName(fd.name, fd)
        self.context[name] = fd
        return fd

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)


class TextFieldDescriptionAddView(FieldDescriptionAddView):

    def create(self, data):
        fd = TextFieldDescription(data['title'],
                                  str(data['name']),
                                  data['required'])
        form.applyChanges(self, fd, data)
        self._fd = fd
        return fd


class DescriptionFieldDescriptionAddView(FieldDescriptionAddView):

    def create(self, data):
        fd = DescriptionFieldDescription(data['title'],
                                         str(data['name']),
                                         data['required'])
        form.applyChanges(self, fd, data)
        self._fd = fd
        return fd


class DateFieldDescriptionAddView(FieldDescriptionAddView):

    def create(self, data):
        fd = DateFieldDescription(data['title'],
                                  str(data['name']),
                                  data['required'])
        form.applyChanges(self, fd, data)
        self._fd = fd
        return fd


class BoolFieldDescriptionAddView(FieldDescriptionAddView):

    def create(self, data):
        fd = BoolFieldDescription(data['title'],
                                  str(data['name']),
                                  data['required'])
        form.applyChanges(self, fd, data)
        self._fd = fd
        return fd


class EnumFieldDescriptionAddView(FieldDescriptionAddView):

    fields = field.Fields(IEnumFieldDescription).select('title', 'name',
        'items', 'required', 'limit_keys')

    def create(self, data):
        fd = EnumFieldDescription(data['title'],
                                  str(data['name']),
                                  data['required'])
        form.applyChanges(self, fd, data)
        self._fd = fd
        return fd


class IntFieldDescriptionAddView(FieldDescriptionAddView):

    fields = field.Fields(IIntFieldDescription).select('title', 'name',
        'min_value', 'max_value', 'required', 'limit_keys')

    def create(self, data):
        fd = IntFieldDescription(data['title'],
                                 str(data['name']),
                                 data['required'])
        form.applyChanges(self, fd, data)
        self._fd = fd
        return fd


class FieldDescriptionEditView(form.EditForm):
    """Edit form for basic person."""
    form.extends(form.EditForm)
    template = ViewPageTemplateFile('templates/person_add.pt')
    fields = field.Fields(IFieldDescription).omit('name')

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(FieldDescriptionEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @property
    def label(self):
        return _(u'Change information for ${field_title}',
                 mapping={'field_title': self.context.title})


class EnumFieldDescriptionEditView(FieldDescriptionEditView):

    fields = field.Fields(IEnumFieldDescription).select('title',
        'items', 'required', 'limit_keys')


class IntFieldDescriptionEditView(FieldDescriptionEditView):

    fields = field.Fields(IIntFieldDescription).select('title',
        'min_value', 'max_value', 'required', 'limit_keys')


class FieldDescriptionView(form.DisplayForm):
    """Display form for a field description."""
    template = ViewPageTemplateFile('templates/field_description_view.pt')
    fields = field.Fields(IFieldDescription)

    def __call__(self):
        self.update()
        return self.render()


class TextFieldDescriptionView(FieldDescriptionView):
    """Display form for a text field description."""


class DescriptionFieldDescriptionView(FieldDescriptionView):
    """Display form for a description field description."""


class DateFieldDescriptionView(FieldDescriptionView):
    """Display form for a date field description."""


class BoolFieldDescriptionView(FieldDescriptionView):
    """Display form for a bool field description."""


class EnumFieldDescriptionView(FieldDescriptionView):
    """Display form for an enum field description."""

    fields = field.Fields(IEnumFieldDescription)


class IntFieldDescriptionView(FieldDescriptionView):
    """Display form for an int field description."""

    fields = field.Fields(IIntFieldDescription)


class FlourishDemographicsFieldsLinks(flourish.page.RefineLinksViewlet):
    """demographics fields add links viewlet."""


class FlourishDemographicsFieldsActions(flourish.page.RefineLinksViewlet):
    """demographics fields action links viewlet."""


class FlourishDemographicsView(flourish.page.WideContainerPage):

    keys = [('students', _("Stud.")),
            ('teachers', _("Teach.")),
            ('administrators', _("Admin."))]

    types = {
        'Enum': _('Selection'),
        'Date': _('Date'),
        'Text': _('Text'),
        'Bool': _('Yes/No'),
        'Description': _('Description'),
        'Int': _('Integer'),
        }

    def table(self):
        result = []
        for demo in list(self.context.values()):
            classname = demo.__class__.__name__
            class_key = classname[:classname.find('FieldDescription')]
            field_type = self.types.get(class_key, '')
            limit_keys = demo.limit_keys
            if limit_keys is None:
                limit_keys = []
            result.append({
               'title': demo.title,
               'url': '%s/edit.html' % absoluteURL(demo, self.request),
               'id': demo.name,
               'type': field_type,
               'required': demo.required,
               'limited': bool(limit_keys),
               'groups': [(key[0] in limit_keys)
                          for key in self.keys],
               })
        return result

    def groups(self):
        return [key[1] for key in self.keys]


class FlourishReorderDemographicsView(flourish.page.Page, DemographicsView):

    def demographics(self):
        pos = 0
        result = []
        for demo in self.context.values():
            pos += 1
            result.append({
                    'name': demo.__name__,
                   'title': demo.title,
                   'pos': pos,
                    })
        return result

    def update(self):
        if 'DONE' in self.request:
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)
        elif 'form-submitted' in self.request:
            for demo in self.context.values():
                name = 'delete.%s' % demo.__name__
                if name in self.request:
                    del self.context[demo.__name__]
                    return
            old_pos, new_pos, move_detected = 0, 0, False
            for demo in self.context.values():
                old_pos += 1
                name = getName(demo)
                if 'pos.'+name not in self.request:
                    continue
                new_pos = int(self.request['pos.'+name])
                if new_pos != old_pos:
                    move_detected = True
                    break
            old_pos, new_pos = old_pos-1, new_pos-1
            keys = list(self.context.keys())
            moving = keys[old_pos]
            keys.remove(moving)
            keys.insert(new_pos,moving)
            self.context.updateOrder(keys)


@adapter(IDemographicsFields)
@implementer(IAddEditViewTitle)
def getAddEditViewTitle(context):
    return _('Demographics')


@adapter(IDemographicsFields)
@implementer(ILimitKeysLabel)
def getLimitKeysLabel(context):
    return _('Limit to group(s)')


@adapter(IDemographicsFields)
@implementer(ILimitKeysHint)
def getLimitKeysHint(context):
    return _(u"If you select one or more groups below, this field "
              "will only be displayed in forms and reports for "
              "members of the selected groups.")


class FlourishFieldDescriptionAddView(flourish.form.AddForm,
                                      FieldDescriptionAddView):

    template = InheritTemplate(flourish.page.Page.template)
    label = None
    legend = _('Field Details')
    formErrorsMessage = _('Please correct the marked fields below.')

    @property
    def title(self):
        return getAdapter(self.context, IAddEditViewTitle)

    @button.buttonAndHandler(_('Submit'), name='add')
    def handleAdd(self, action):
        super(FlourishFieldDescriptionAddView, self).handleAdd.func(self, action)

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())

    def updateWidgets(self):
        super(FlourishFieldDescriptionAddView, self).updateWidgets()
        self.widgets['limit_keys'].label = getAdapter(self.context,
            ILimitKeysLabel)
        self.widgets['limit_keys'].field.description = getAdapter(self.context,
            ILimitKeysHint)

    def nextURL(self):
        return absoluteURL(self.context, self.request)


class FlourishTextFieldDescriptionAddView(FlourishFieldDescriptionAddView,
                                          TextFieldDescriptionAddView):
    pass


class FlourishDescriptionFieldDescriptionAddView(FlourishFieldDescriptionAddView,
                                                 DescriptionFieldDescriptionAddView):
    pass


class FlourishDateFieldDescriptionAddView(FlourishFieldDescriptionAddView,
                                          DateFieldDescriptionAddView):
    pass


class FlourishBoolFieldDescriptionAddView(FlourishFieldDescriptionAddView,
                                          BoolFieldDescriptionAddView):
    pass


class FlourishEnumFieldDescriptionAddView(FlourishFieldDescriptionAddView,
                                          EnumFieldDescriptionAddView):
    pass


class FlourishIntFieldDescriptionAddView(FlourishFieldDescriptionAddView,
                                         IntFieldDescriptionAddView):
    pass


class FlourishFieldDescriptionEditView(flourish.form.Form,
                                       FieldDescriptionEditView):

    template = InheritTemplate(flourish.page.Page.template)
    label = None
    legend = _('Field Details')

    @property
    def title(self):
        return getAdapter(self.context.__parent__, IAddEditViewTitle)

    @property
    def subtitle(self):
        return _(u'Change information for ${field_title}',
                 mapping={'field_title': self.context.title})

    @button.buttonAndHandler(_('Submit'), name='apply')
    def handleApply(self, action):
        super(FlourishFieldDescriptionEditView, self).handleApply.func(self, action)
        # XXX: hacky sucessful submit check
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())

    def updateWidgets(self):
        super(FlourishFieldDescriptionEditView, self).updateWidgets()
        self.widgets['limit_keys'].label = getAdapter(
            self.context.__parent__, ILimitKeysLabel)
        self.widgets['limit_keys'].field.description = getAdapter(
            self.context.__parent__, ILimitKeysHint)

    def nextURL(self):
        return absoluteURL(self.context.__parent__, self.request)


class FlourishEnumFieldDescriptionEditView(FlourishFieldDescriptionEditView, EnumFieldDescriptionEditView):

    def update(self):
        EnumFieldDescriptionEditView.update(self)


class FlourishIntFieldDescriptionEditView(FlourishFieldDescriptionEditView, IntFieldDescriptionEditView):

    def update(self):
        IntFieldDescriptionEditView.update(self)


class FlourishTextFieldDescriptionView(flourish.page.Page, TextFieldDescriptionView):

    def update(self):
        TextFieldDescriptionView.update(self)


class FlourishDescriptionFieldDescriptionView(flourish.page.Page,
                                              DescriptionFieldDescriptionView):

    def update(self):
        DescriptionFieldDescriptionView.update(self)


class FlourishDateFieldDescriptionView(flourish.page.Page, DateFieldDescriptionView):

    def update(self):
        DateFieldDescriptionView.update(self)


class FlourishBoolFieldDescriptionView(flourish.page.Page, BoolFieldDescriptionView):

    def update(self):
        BoolFieldDescriptionView.update(self)


class FlourishEnumFieldDescriptionView(flourish.page.Page, EnumFieldDescriptionView):

    def update(self):
        EnumFieldDescriptionView.update(self)


class FlourishIntFieldDescriptionView(flourish.page.Page, IntFieldDescriptionView):

    def update(self):
        IntFieldDescriptionView.update(self)


class InvalidFieldID(ValidationError):
    __doc__ = _(u'Please enter a valid ID')


class DuplicatedFieldID(ValidationError):
    __doc__ = _(u'This ID is not unique.')


class DemographicsFieldIDValidator(validator.SimpleFieldValidator):

    def validate(self, value):
        try:
            super(DemographicsFieldIDValidator, self).validate(value)
        except (ValidationError,):
            raise InvalidFieldID()
        if value in self.context:
            raise DuplicatedFieldID()


validator.WidgetValidatorDiscriminators(DemographicsFieldIDValidator,
                                        field=IFieldDescription['name'])



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Membership relationship.

This module defines group membership as a relationship.

We can reuse most of the test fixture from schooltool.relationship.tests.

    >>> from schooltool.relationship.tests import setUp, tearDown
    >>> setUp()

We also need to register an event subscriber for IBeforeRelationshipEvent,
to enforce relationship constraints.

    >>> import zope.event
    >>> old_subscribers = zope.event.subscribers[:]
    >>> zope.event.subscribers.append(enforceMembershipConstraints)

We will need some sample persons and groups for the demonstration

    >>> from schooltool.group.group import Group
    >>> from schooltool.person.person import Person
    >>> jonas = Person()
    >>> petras = Person()
    >>> developers = Group()
    >>> admins = Group()

You can create memberships this way:

    >>> from schooltool.app.membership import Membership
    >>> Membership(member=jonas, group=developers)
    >>> Membership(member=petras, group=developers)
    >>> Membership(member=petras, group=admins)

You can find all members of a group, or groups of a persons this way:

    >>> from schooltool.relationship import getRelatedObjects
    >>> from schooltool.app.membership import URIMember, URIGroup
    >>> set(getRelatedObjects(developers, URIMember)) == set([jonas, petras])
    True
    >>> set(getRelatedObjects(petras, URIGroup)) == set([admins, developers])
    True

There are some constraints: Only objects providing IGroup can be groups.

    >>> Membership(member=jonas, group=petras)
    Traceback (most recent call last):
      ...
    InvalidRelationship: Groups must provide IGroup.

You may not create cyclic memberships as support is dropped now for
group-to-group Membership relationships.

    >>> Membership(member=admins, group=developers)
    Traceback (most recent call last):
      ...
    InvalidRelationship: Groups cannot be members of a group anymore.

You may not create ill-formed relationships

    >>> from schooltool.relationship import relate
    >>> from schooltool.app.membership import URIMembership
    >>> relate(URIMembership, (jonas, URIMember), (petras, URIMember))
    Traceback (most recent call last):
      ...
    InvalidRelationship: Membership must have one member and one group.

    >>> relate(URIMembership, (jonas, URIGroup), (petras, URIGroup))
    Traceback (most recent call last):
      ...
    InvalidRelationship: Membership must have one member and one group.

Resources can't be part of a group:

    >>> from schooltool.resource.resource import Resource
    >>> printer = Resource()
    >>> relate(URIMembership, (admins, URIGroup), (printer, URIMember))
    Traceback (most recent call last):
      ...
    InvalidRelationship: Resources cannot be members of a group.

Of course, these constraints do not apply to other kinds of relationships.

    >>> from schooltool.relationship.tests import URIStub
    >>> relate('example:Frogship', (jonas, URIStub('example:Frog')),
    ...                            (petras, URIStub('example:Frog')))

That's all.

    >>> zope.event.subscribers[:] = old_subscribers
    >>> tearDown()

"""

from zope.component import adapts

from schooltool.relationship import URIObject, RelationshipSchema
from schooltool.relationship import getRelatedObjects
from schooltool.relationship.interfaces import IBeforeRelationshipEvent
from schooltool.relationship.interfaces import InvalidRelationship
from schooltool.resource.interfaces import IBaseResource
from schooltool.group.interfaces import IBaseGroup as IGroup
from schooltool.person.interfaces import IPerson
from schooltool.securitypolicy.crowds import Crowd

from schooltool.common import SchoolToolMessage as _


URIMembership = URIObject('http://schooltool.org/ns/membership',
                          'Membership', 'The membership relationship.')
URIGroup = URIObject('http://schooltool.org/ns/membership/group',
                     'Group', 'A role of a containing group.')
URIMember = URIObject('http://schooltool.org/ns/membership/member',
                      'Member', 'A group member role.')

Membership = RelationshipSchema(URIMembership,
                                member=URIMember,
                                group=URIGroup)


def enforceMembershipConstraints(event):
    """Enforce membership constraints (IBeforeRelationshipEvent subscriber)."""
    # XXX look through all the IBeforeRelationshipEvent subscribers,
    # if there are any, and convert them to something different.
    if not IBeforeRelationshipEvent.providedBy(event):
        return
    if event.rel_type != URIMembership:
        return
    if ((event.role1, event.role2) != (URIMember, URIGroup) and
        (event.role1, event.role2) != (URIGroup, URIMember)):
        raise InvalidRelationship('Membership must have one member'
                                  ' and one group.')
    if IBaseResource.providedBy(event[URIMember]):
        raise InvalidRelationship("Resources cannot be members of a group.")
    if not IGroup.providedBy(event[URIGroup]):
        raise InvalidRelationship('Groups must provide IGroup.')
    if IGroup.providedBy(event[URIMember]):
        raise InvalidRelationship("Groups cannot be members of a group anymore.")
    if isTransitiveMember(event[URIGroup], event[URIMember]):
        raise InvalidRelationship('No cycles are allowed.')


def isTransitiveMember(obj, group):
    """Is `obj` a member of `group` (either directly or indirectly)?

        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> setUp()

    Suppose we have four groups, named a, b, c and d.

        >>> from schooltool.group.group import Group
        >>> a, b, c, d = [Group() for n in range(4)]

    A is a member of b, which is a member of c.

        >>> Membership(member=a, group=b)
        >>> Membership(member=b, group=c)

    If a is a member of b, it is, obviously, a transitive member

        >>> isTransitiveMember(a, b)
        True

    If a is a member of b, which is a member of c, then a is a transitive
    member of c.

        >>> isTransitiveMember(a, c)
        True

    We say that a is a transitive member of a, for convenience.

        >>> isTransitiveMember(a, a)
        True

    In all other cases, we return False.

        >>> isTransitiveMember(b, a)
        False
        >>> isTransitiveMember(c, a)
        False
        >>> isTransitiveMember(a, d)
        False

    That's it.

        >>> tearDown()

    """
    if obj is group:
        return True
    # A group usually has more members than a member has groups, so we will
    # find all transitive groups of `obj` and see whether `group` is one of
    # them.  It does not matter if we use breadth-first or depth-first search.
    queue = [obj]
    seen = set([id(obj)])
    while queue:
        cur_obj = queue.pop() # therefore DFS; change to pop(0) to get BFS
        for new_group in getRelatedObjects(cur_obj, URIGroup):
            if new_group is group:
                return True
            if id(new_group) not in seen:
                seen.add(id(new_group))
                queue.append(new_group)
    return False


class GroupMemberCrowd(Crowd):
    """Crowd that contains all the members of the group.

        >>> from schooltool.app.membership import GroupMemberCrowd
        >>> class GroupStub(object):
        ...     def __init__(self, members):
        ...         self.members = members
        >>> class PrincipalStub(object):
        ...     def __init__(self, name):
        ...         self.name = name
        ...     def __conform__(self, iface):
        ...         if iface == IPerson:
        ...             return self.name
        >>> group = GroupStub(["Petras"])
        >>> crowd = GroupMemberCrowd(group)

    If person is not a member of the group return False:

        >>> crowd.contains(PrincipalStub("Jonas"))
        False

    If the person is in the member list of the context group return
    True:

        >>> crowd.contains(PrincipalStub("Petras"))
        True

    Some principals might not adapt to IPerson, they should be
    rejected:

        >>> crowd.contains("Foo")
        False

    """

    adapts(IGroup)

    title = _(u'Members')
    description = _(u'Members of the group.')

    def contains(self, principal):
        return IPerson(principal, None) in self.context.members



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Meta directives and their handlers.
"""
import zope.component.zcml
import zope.configuration.fields
from zope.interface import Interface
from zope.security.zcml import Permission

from schooltool.skin.flourish.breadcrumbs import IBreadcrumb
from schooltool.skin.flourish.breadcrumbs import Breadcrumbs
from schooltool.skin.flourish.breadcrumbs import PageBreadcrumbs
from schooltool.skin.flourish import interfaces
from schooltool.skin.flourish.zcml import IContentOrientedDirective
from schooltool.skin.flourish.zcml_content import handle_security
from schooltool.skin.flourish.zcml_content import IContentDirective
from schooltool.skin.flourish.zcml_content import contentDirective


class IBreadcrumbDirectiveBase(Interface):

    crumb_parent = zope.configuration.fields.GlobalObject(
        title=u"Crumb parent",
        description=u"Getter of the crumb parent.",
        required=False,
        )

    follow_crumb = zope.configuration.fields.GlobalObject(
        title=u"Follow crumb",
        description=u"The breadcrumb factory.",
        required=False,
        )


class IBreadcrumbDirective(IContentOrientedDirective,
                           IBreadcrumbDirectiveBase):
    """Define the SchoolTool content provider."""

    title = zope.configuration.fields.MessageID(
        title=u"Title",
        required=False,
        )

    class_ = zope.configuration.fields.GlobalObject(
        title=u"Class",
        description=u"The breadcrumb view class.",
        required=False,
        default=Breadcrumbs,
        )


IBreadcrumbDirective.setTaggedValue('keyword_arguments', True)


class IPageBreadcrumbsDirective(IContentDirective):
    """Define breadcrumbs content for a page."""

    name = zope.schema.TextLine(
        title=u"The name of the content provider.",
        required=False,
        default=u'breadcrumbs')

    permission = Permission(
        title=u"Permission",
        description=u"The permission needed to use the view.",
        required=False,
        )

    crumb_parent = zope.configuration.fields.GlobalObject(
        title=u"Crumb parent",
        description=u"Getter of the crumb parent.",
        required=False,
        )

    follow_crumb = zope.configuration.fields.GlobalObject(
        title=u"Follow crumb",
        description=u"The breadcrumb factory.",
        required=False,
        )

    show_page_title = zope.configuration.fields.Bool(
        title=u"Show page title",
        required=False,
        default=True)


IPageBreadcrumbsDirective.setTaggedValue('keyword_arguments', True)


def breadcrumbDirective(
    _context,
    for_=Interface, layer=interfaces.IFlourishLayer, view=interfaces.IPage,
    crumb_parent=None, follow_crumb=None,
    class_=Breadcrumbs,
    **kwargs):

    class_dict = dict(kwargs)

    bases = (class_, )
    if not IBreadcrumb.implementedBy(class_):
        bases = bases + (Breadcrumbs, )

    if crumb_parent is not None:
        class_dict['crumb_parent'] = property(
            lambda crumb: crumb_parent(crumb.context))

    if follow_crumb is not None:
        class_dict['follow_crumb'] = property(
            lambda crumb: follow_crumb(
                crumb.crumb_parent, crumb.request, crumb.view))

    class_ = type(class_.__name__, bases, class_dict)

    # XXX: I'm pretty sure security is not handled correctly here
    handle_security(class_, 'zope.Public', IBreadcrumb, ())

    _context.action(
        discriminator=('schooltool.skin.flourish.breadcrumb',
                       for_, layer, view),
        callable=zope.component.zcml.handler,
        args=('registerAdapter',
              class_,
              (for_, layer, view),
              IBreadcrumb,
              '',
              _context.info),
        )


def pageBreadcrumbsDirective(
    _context, name=u'breadcrumbs', permission=u'zope.Public',
    for_=Interface, layer=interfaces.IFlourishLayer, view=interfaces.IPage,
    class_=None, show_page_title=True,
    **kwargs):

    if class_ is None:
        if show_page_title:
            class_ = PageBreadcrumbs
        else:
            class_ = Breadcrumbs

    kwargs = dict(kwargs)
    _crumb_parent_factory = kwargs.get('crumb_parent')
    _follow_crumb_factory = kwargs.get('follow_crumb')

    if _crumb_parent_factory is not None:
        kwargs['crumb_parent'] = property(
            lambda crumb: _crumb_parent_factory(crumb.context))

    if _follow_crumb_factory is not None:
        kwargs['follow_crumb'] = property(
            lambda crumb: _follow_crumb_factory(
                crumb.crumb_parent, crumb.request, crumb.view))

    contentDirective(
        _context, name, permission,
        for_=for_, layer=layer, view=view,
        class_=class_, **kwargs)
